from earth_extractor.providers import Provider
import logging
from earth_extractor.satellites import enums
from earth_extractor import core
from earth_extractor.core.credentials import get_credentials
from earth_extractor.core.models import CommonSearchResult
from typing import Any, List, TYPE_CHECKING, Optional
import pyproj
from shapely.ops import transform
import datetime
import shapely.geometry
import requests


if TYPE_CHECKING:
    from earth_extractor.satellites.base import Satellite

credentials = get_credentials()

logger = logging.getLogger(__name__)
logger.setLevel(core.config.constants.LOGLEVEL_MODULE_DEFAULT)


class SwissTopo(Provider):
    def query(
        self,
        satellite: "Satellite",
        processing_level: enums.ProcessingLevel,
        roi: shapely.geometry.base.BaseGeometry,
        start_date: datetime.datetime,
        end_date: datetime.datetime,
        cloud_cover: Optional[int] = None,
    ) -> List[CommonSearchResult]:
        """Perform a search query as if done via SwissTopo web interface

        Method discovery by inspecting the network traffic of the SwissTopo
        web interface.
        https://www.swisstopo.admin.ch/en/geodata/images/ortho/swissimage10.html#download  # noqa: E501

        Local ROI is WGS84, EPSG:4326, but SwissTopo API requires CH1903+
        coordinates. We must cast the ROI to CH1903+ EPSG: 2056 before querying
        the API even though there appears to be a CRS parameter in the
        API -- as API is undocumented, it's best to play it safe.
        """

        logger.info(f"{self.name}: Querying API")

        # Convert datetime to unix time 1601942400000 represents 2020-10-06
        # two hours after midnight (Zurich time) representing an offset of 7200
        # seconds. Therefore an incoming date of 2020-10-06 will be converted
        # to 1601942400000. The unix timestamp here is also represented in
        # milliseconds, so we multiply by 1000.
        start_date_unix = int(
            (start_date + datetime.timedelta(hours=2)).timestamp() * 1000
        )

        # I have a suspicion that start time is offset by +2 hours, to
        # Europe/Zurich but the end_time is still UTC +0 hours. There is no
        # adjustment here for this, but it may be worth investigating in the
        # future.
        end_date_unix = int(
            (end_date + datetime.timedelta(hours=2)).timestamp() * 1000
        )
        request_epsg = 2056

        # Get the resolution from the products dictionary based on CLI choice
        # Normally this would be a satellite processing level, but in this case
        # it's a resolution. Rather than redesigning the code for this one use
        # case (SwissTopo), we can just use the resolution as the processing
        # level. May be worth changing in the future
        resolution = self.products[(satellite.name, processing_level)][0]
        product = "ch.swisstopo.swissimage-dop10"

        # Reproject coordinates from WGS84 to CH1903+ (EPSG:2056)
        wgs84 = pyproj.CRS("EPSG:4326")
        utm = pyproj.CRS("EPSG:2056")

        project = pyproj.Transformer.from_crs(
            wgs84, utm, always_xy=True
        ).transform
        roi_utm = transform(project, roi)
        x_min, y_min, x_max, y_max = roi_utm.bounds

        # Options seem to be 2017, 2018, 2019, 2020, 2021, 2022 and "current".
        # If set to None, it will return all years which is confusing because
        # we can also query for time period. We can do a string filter
        # on the results to only return the years we want (faster queries but
        # a bit hacky). Ensure the assumptions are tested at runtime.
        year_collection = None

        url = (
            "https://ogd.swisstopo.admin.ch/services/swiseld/services/assets/"
            f"{product}/search"
            "?format=image/tiff; application=geotiff; "
            f"profile=cloud-optimized&{resolution}&srid={request_epsg}"
            f"&state={year_collection if year_collection is not None else ''}"
            f"&from={start_date_unix}&to={end_date_unix}"
            f"&xMin={x_min}&yMin={y_min}&xMax={x_max}&yMax={y_max}&csv=true"
        )

        # It is possible to get the geometry via this request, but the data
        # does not contain any links to crossreference to the download link.
        # It is also possible to perform a query without geometry_only and csv
        # to be given a list of links with geometries, but there is a server
        # limit set that if you hit it, you are returned with no data
        # and limit=True in the response. Therefore, it is a balance of either
        # getting the geometry and no links, or links and no geometry. The
        # latter is preferred as we can at least download the data.

        # url = (
        #     "https://ogd.swisstopo.admin.ch/services/swiseld/services/assets/"
        #     f"{product}/search"
        #     "?format=image/tiff; application=geotiff; "
        #     f"profile=cloud-optimized&{resolution}&srid={request_epsg}"
        #     f"&state={year_collection if year_collection is not None else ''}"
        #     f"&from={start_date_unix}&to={end_date_unix}"
        #     f"&xMin={x_min}&yMin={y_min}&xMax={x_max}&yMax={y_max}"
        #     "&geometry_only=true"
        # )

        res = requests.get(url)
        res.raise_for_status()
        logger.debug(f"Unix time: {start_date_unix}, {end_date_unix}")
        logger.debug(f"Bounds {x_min}, {y_min}, {x_max}, {y_max}")

        # Get the CSV path from JSON response
        csv_path = res.json()["href"]
        logger.debug(f"CSV Path: {csv_path}")

        # import csv
        res = requests.get(csv_path)
        res.raise_for_status()

        # The "csv" is a headerless, single-column CSV that can be parsed into
        # one line per file by splitting by newline char
        file_list = res.content.decode("utf-8").split("\n")

        if len(file_list) == 0 or (len(file_list) == 1 and file_list[0] == ""):
            logger.info(f"{self.name}: No results found for query")
            return []

        # We can try and make an assumption that the returned addresses stated:
        # https://<url>/swissimage-dop10_<i>_<j>-<k>/filename.tif can relate to
        # the following:
        # i = year
        # j = unknown (flight #?)
        # k = unknown (tile #?)
        # We can then filter the results based on the year we want to download
        # and then download the files.
        filtered_file_list = []
        for url_path in file_list:
            subpath, _filename = url_path.split("/")[-2:]

            # Split the subpath into its components (as listed above)
            subpath_split = subpath.split("_")
            if (
                subpath_split[0] == "swissimage-dop10"
                and len(str(subpath_split[1])) == 4
            ):
                # Get the range of years given by the user in a sequential list
                # and then filter the results based on that list
                year_range = list(range(start_date.year, end_date.year + 1))
                if int(subpath_split[1]) in year_range:
                    filtered_file_list.append(url_path)
            else:
                raise ValueError(
                    f"Unexpected file path provided by SwissTopo: '{url_path}'"
                    " This breaks the assumption of yearly filtering. Please "
                    "report this issue to the developers."
                )

        logger.info(
            f"{self.name}: Found {len(filtered_file_list)} files to download"
        )

        records = self.translate_search_results(
            provider_search_results=filtered_file_list,
            resolution=resolution,
            roi=roi,
        )

        return records

    def translate_search_results(
        self,
        *,
        provider_search_results: Any,
        resolution: str,
        roi: shapely.geometry.base.BaseGeometry,
    ) -> List[CommonSearchResult]:
        """Translate search results from a provider to a common format

        As explained above in comments, the SwissTopo API does not provide any
        geometry information. Therefore, we must use the ROI that was used to
        query the API to create a geometry for each result. This is not ideal
        as the geometry is not the exact geometry of the image, but it is the
        best we can do with the information provided by the API.

        If this is resolved at a later date it may be worth removing the
        warning message when doing an export that includes swissimage in
        query.py.
        """

        sat, level = self._products_reversed[resolution]

        common_results = []
        for item in provider_search_results:
            # Swiss Topo does not provide any geometry information

            common_results.append(
                CommonSearchResult(
                    url=item,
                    satellite=sat,
                    geometry=roi.wkt,
                    processing_level=level,
                    notes=(
                        "SwissImage geometry reflects only the given ROI, "
                        "not the actual boundary of the image. See comments "
                        "within the code for more information."
                    ),
                )
            )

        return common_results

    def download_many(
        self,
        search_results: List[CommonSearchResult],
        download_dir: str,
        overwrite: bool = False,
        processes: int = core.config.constants.PARRALLEL_PROCESSES_DEFAULT,
    ) -> None:
        """Download many files from the SwissTopo API

        Parameters
        ----------
        search_results : List[str]
            The search results
        download_dir : str
            The directory to download the data to
        overwrite : bool, optional
            Whether to overwrite existing files, by default False
        processes : int, optional
            The number of processes to use for downloading

        Uses the common download utility with progress bar.
        """

        urls = [str(x.url) for x in search_results]
        core.utils.download_parallel(
            urls, download_dir, overwrite=overwrite, processes=processes
        )


swiss_topo: SwissTopo = SwissTopo(
    name="SwissTopo",
    description="Swiss Federal Office of Topography",
    products={
        (enums.Satellite.SWISSIMAGE, enums.ProcessingLevel.CM10): [
            "resolution=0.1"
        ],
        (enums.Satellite.SWISSIMAGE, enums.ProcessingLevel.CM200): [
            "resolution=2.0"
        ],
    },
    uri="https://data.geo.admin.ch",
)



from pydantic import BaseSettings
import logging
import os
import datetime


class Constants(BaseSettings):
    COMMON_TIMESTAMP: str = (
        # Windows doesn't like colons in filenames
        f"{datetime.datetime.utcnow().isoformat().replace(':', '-')}"
    )
    DEFAULT_DOWNLOAD_DIR: str = os.path.join(os.getcwd(), "data")
    GEOJSON_EXPORT_FILENAME: str = f"{COMMON_TIMESTAMP}.geojson"
    MAX_DOWNLOAD_ATTEMPTS: int = 50
    HIDE_PASSWORD_PROMPT: bool = False

    DEFAULT_DOWNLOAD_THREADS: int = 10
    PARRALLEL_PROCESSES_DEFAULT: int = 4

    KEYRING_ID: str = "earth-extractor"

    # Logging
    LOGLEVEL_MODULE_DEFAULT: int = logging.DEBUG
    LOGFILE_NAME: str = f"{COMMON_TIMESTAMP}.log"
    LOGLEVEL_FILE: int = logging.DEBUG
    LOGLEVEL_CONSOLE: int = logging.INFO
    LOGFORMAT_CONSOLE: logging.Formatter = logging.Formatter(
        "%(asctime)s [%(levelname)-7.7s] %(message)s"
    )
    LOGFORMAT_FILE = logging.Formatter(
        "%(asctime)s [%(levelname)-7.7s] (%(name)25.25s) %(message)s"
    )


constants = Constants()



from earth_extractor.providers import Provider
import logging
from earth_extractor.satellites import enums
from earth_extractor import core
from earth_extractor.core.credentials import get_credentials
from earth_extractor.core.models import CommonSearchResult
from typing import Any, List, TYPE_CHECKING, Optional
import datetime
import shapely.geometry
from shapely.geometry import Polygon


if TYPE_CHECKING:
    from earth_extractor.satellites.base import Satellite

credentials = get_credentials()

logger = logging.getLogger(__name__)
logger.setLevel(core.config.constants.LOGLEVEL_MODULE_DEFAULT)


class NASACommonMetadataRepository(Provider):
    def query(
        self,
        satellite: "Satellite",
        processing_level: enums.ProcessingLevel,
        roi: shapely.geometry.base.BaseGeometry,
        start_date: datetime.datetime,
        end_date: datetime.datetime,
        cloud_cover: Optional[int] = None,
    ) -> List[CommonSearchResult]:
        """Query the NASA Common Metadata Repository with STAC"""

        # Check that the provider's credentials that are needed are set
        self._check_credentials_exist()

        logger.info("Querying NASA Common Metadata Repository")
        res = self.query_stac(
            provider_uri=f"{self.uri}/stac/LAADS",
            collections=self.products[(satellite.name, processing_level)],
            roi=roi,
            start_date=start_date,
            end_date=end_date,
        )
        features = res["features"]

        logger.info(f"NASA CMR: Found {len(features)} files to download")

        return self.translate_search_results(features)

    def translate_search_results(
        self, provider_search_results: Any
    ) -> List[CommonSearchResult]:
        """Translate search results from a provider to a common format"""

        common_results = []
        for record in provider_search_results:
            # Get the satellite and processing level from reversed mapping of
            # the provider's "products" dictionary
            sat, level = self._products_reversed[record["collection"]]
            datetime_obj = datetime.datetime.strptime(
                record["properties"]["datetime"], "%Y-%m-%dT%H:%M:%S.%fZ"
            )

            common_results.append(
                CommonSearchResult(
                    product_id=record["id"],
                    time=datetime_obj,
                    geometry=Polygon(record["geometry"]["coordinates"][0]),
                    url=record["assets"]["data"]["href"],
                    processing_level=level,
                    satellite=sat,
                )
            )

        return common_results

    def download_many(
        self,
        search_results: List[CommonSearchResult],
        download_dir: str,
        overwrite: bool = False,
        processes: int = core.config.constants.PARRALLEL_PROCESSES_DEFAULT,
    ) -> None:
        """Using the search results from query(), download the data

        Parameters
        ----------
        search_results : List[str]
            The search results
        download_dir : str
            The directory to download the data to
        overwrite : bool, optional
            Whether to overwrite existing files, by default False
        processes : int, optional
            The number of processes to use for downloading

        Returns
        -------
        None
        """

        # Convert the search results to a list of URIs

        # Check that the provider's credentials that are needed are set
        self._check_credentials_exist()

        urls = [str(x.url) for x in search_results if x.url]

        auth_header = {"Authorization": f"Bearer {credentials.NASA_TOKEN}"}

        for url in urls:
            try:
                core.utils.download_parallel(
                    urls, download_dir, auth_header, overwrite, processes
                )
            except RuntimeError as e:
                # Log the exception and continue to the next file
                logger.error(e)
                continue


nasa_cmr: NASACommonMetadataRepository = NASACommonMetadataRepository(
    name="CMR",
    description="NASA Common Metadata Repository",
    products={
        (enums.Satellite.MODIS_TERRA, enums.ProcessingLevel.L1B): [
            "MOD02QKM.v6.1",
            "MOD02HKM.v6.1",
            "MOD021KM.v6.1",
            "MOD03.v6.1",
        ],
        (enums.Satellite.MODIS_AQUA, enums.ProcessingLevel.L1B): [
            "MYD02QKM.v6.1",
            "MYD02HKM.v6.1",
            "MYD021KM.v6.1",
            "MYD03.v6.1",
        ],
        (enums.Satellite.VIIRS, enums.ProcessingLevel.L1): [
            "VNP03IMG.v2",
            "VJ103IMG.v2.1",
            "VNP03MOD.v2",
            "VJ103MOD.v2.1",
        ],
    },
    uri="https://cmr.earthdata.nasa.gov",
    credentials_required=["NASA_TOKEN"],
)



import logging
from earth_extractor import core
import os


def setup_file_logger(output_folder):
    # Add a file logger

    # Create folder if it doesn't exist
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)

    logfile_handler = logging.FileHandler(
        os.path.join(output_folder, core.constants.LOGFILE_NAME)
    )
    logfile_handler.setLevel(core.constants.LOGLEVEL_FILE)
    logfile_handler.setFormatter(core.constants.LOGFORMAT_FILE)
    logging.getLogger().addHandler(logfile_handler)


class ErrorFlagHandler(logging.Handler):
    """Catches ERRORs to allow app to know if an error occurred at exit"""

    def __init__(self):
        super().__init__()
        self.errors = 0

    def emit(self, record):
        if record.levelno >= logging.ERROR:
            self.errors += 1

    def print_status(self):
        if self.errors > 0:
            print()
            print(
                f"There {'were' if self.errors > 1 else 'was'} {self.errors} "
                f"error{'s' if self.errors > 1 else ''} during the "
                "execution of the app. Please check the error log file "
                f"'{core.config.constants.LOGFILE_NAME}' for more details."
            )



from enum import Enum


class ProcessingLevel(str, Enum):
    """Satellite processing levels
    As defined by:
    https://www.earthdata.nasa.gov/engage/open-data-services-and-software/data-information-policy/data-levels
    """

    L0 = "L0"
    L1 = "L1"
    L1A = "L1A"
    L1B = "L1B"
    L1C = "L1C"
    L2 = "L2"
    L2A = "L2A"
    L2B = "L2B"
    L3 = "L3"
    L3A = "L3A"
    L4 = "L4"

    # Spatial Resolutions (breaks the naming convention unfortunately ...)
    # SWISSIMAGE is provided in two products, 0.1m and 2m
    # Perhaps reassess naming of "ProcessingLevel" to something else later if
    # resolutions in Satellite clases are to be added
    CM10 = "CM10"
    CM200 = "CM200"


class Sensor(str, Enum):
    MSI = "MSI"
    C_SAR = "C-SAR"
    OLCI = "OLCI"
    SLSTR = "SLSTR"
    SRAL = "SRAL"
    MODIS = "MODIS"
    VIIRS = "VIIRS"
    SWISSIMAGE = "SWISSIMAGE"


class Satellite(Enum):
    """Contains the supported satellites supported by the library"""

    SENTINEL1 = "SENTINEL1"
    SENTINEL2 = "SENTINEL2"
    SENTINEL3 = "SENTINEL3"
    MODIS_TERRA = "MODIS_TERRA"
    MODIS_AQUA = "MODIS_AQUA"
    VIIRS = "VIIRS"
    SWISSIMAGE = "SWISSIMAGE_10CM"


class Filters(Enum):
    """The available filters for the satellites

    To understand the filtering capabilities for each sensor, for example
    radar satellites do not have a cloud cover percentage, so manage those that
    do have it.
    """

    CLOUD_COVER = "CLOUD_COVER"



""" Functions from the sentinelsat library, modified to allow overwriting of
existing files.

Full repository: https://github.com/sentinelsat/sentinelsat
"""


from sentinelsat import SentinelAPI, Downloader, DownloadStatus
from sentinelsat.download import _format_exception
from collections import namedtuple
import threading
from concurrent.futures import ThreadPoolExecutor
import concurrent.futures
import itertools
from sentinelsat.exceptions import (
    SentinelAPIError,
)
from copy import copy


class SentinelAPIExtended(SentinelAPI):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.downloader = DownloaderExtended(self)  # @evanjt: Add ext. class

    def download_all(
        self,
        products,
        directory_path=".",
        max_attempts=10,
        checksum=True,
        n_concurrent_dl=None,
        lta_retry_delay=None,
        fail_fast=False,
        nodefilter=None,
        overwrite=False,
    ):
        """Download a list of products.

        Takes a list of product IDs as input. This means that the return value of query() can be
        passed directly to this method.

        File names on the server are used for the downloaded files, e.g.
        "S1A_EW_GRDH_1SDH_20141003T003840_20141003T003920_002658_002F54_4DD1.zip".

        In case of interruptions or other exceptions, downloading will restart from where it left
        off. Downloading is attempted at most max_attempts times to avoid getting stuck with
        unrecoverable errors.

        Parameters
        ----------
        products : list
            List of product IDs
        directory_path : string
            Directory where the downloaded files will be downloaded
        max_attempts : int, default 10
            Number of allowed retries before giving up downloading a product.
        checksum : bool, default True
            If True, verify the downloaded files' integrity by checking its MD5 checksum.
            Throws InvalidChecksumError if the checksum does not match.
            Defaults to True.
        n_concurrent_dl : integer, optional
            Number of concurrent downloads. Defaults to :attr:`SentinelAPI.concurrent_dl_limit`.
        lta_retry_delay : float, default 60
            Number of seconds to wait between requests to the Long Term Archive.
        fail_fast : bool, default False
            if True, all other downloads are cancelled when one of the downloads fails.
        nodefilter : callable, optional
            The callable is used to select which files of each product will be downloaded.
            If None (the default), the full products will be downloaded.
            See :mod:`sentinelsat.products` for sample node filters.
        overwrite: bool, default False
            If True, existing files will be overwritten. If False, existing files will be skipped.
        Notes
        -----
        By default, raises the most recent downloading exception if all downloads failed.
        If ``fail_fast`` is set to True, raises the encountered exception on the first failed
        download instead.

        Returns
        -------
        dict[string, dict]
            A dictionary containing the return value from download() for each successfully
            downloaded product.
        dict[string, dict]
            A dictionary containing the product information for products successfully
            triggered for retrieval from the long term archive but not downloaded.
        dict[string, dict]
            A dictionary containing the product information of products where either
            downloading or triggering failed. "exception" field with the exception info
            is included to the product info dict.
        """
        downloader = copy(self.downloader)
        downloader.verify_checksum = checksum
        downloader.fail_fast = fail_fast
        downloader.max_attempts = max_attempts
        if n_concurrent_dl:
            downloader.n_concurrent_dl = n_concurrent_dl
        if lta_retry_delay:
            downloader.lta_retry_delay = lta_retry_delay
        downloader.node_filter = nodefilter
        statuses, exceptions, product_infos = downloader.download_all(
            products, directory_path, overwrite=overwrite
        )

        # Adapt results to the old download_all() API
        downloaded_prods = {}
        retrieval_triggered = {}
        failed_prods = {}
        for pid, status in statuses.items():
            if pid not in product_infos:
                product_infos[pid] = {}
            if pid in exceptions:
                product_infos[pid]["exception"] = exceptions[pid]
            if status == DownloadStatus.DOWNLOADED:
                downloaded_prods[pid] = product_infos[pid]
            elif status == DownloadStatus.TRIGGERED:
                retrieval_triggered[pid] = product_infos[pid]
            else:
                failed_prods[pid] = product_infos[pid]
        ResultTuple = namedtuple(
            "ResultTuple", ["downloaded", "retrieval_triggered", "failed"]
        )
        return ResultTuple(downloaded_prods, retrieval_triggered, failed_prods)


class DownloaderExtended(Downloader):
    def download_all(self, products, directory=".", overwrite=False):
        """Download a list of products.

        Parameters
        ----------
        products : list
            List of product IDs
        directory : string or Path, optional
            Directory where the files will be downloaded

        Notes
        ------
        By default, raises the most recent downloading exception if all downloads failed.
        If :attr:`Downloader.fail_fast` is set to True, raises the encountered exception on the first failed
        download instead.

        Returns
        -------
        dict[string, DownloadStatus]
            The status of all products.
        dict[string, Exception]
            Exception info for any failed products.
        dict[string, dict]
            A dictionary containing the product information for each product
            (unless the product was unavailable).
        """

        ResultTuple = namedtuple(
            "ResultTuple", ["statuses", "exceptions", "product_infos"]
        )
        product_ids = list(set(products))
        assert self.n_concurrent_dl > 0
        if len(product_ids) == 0:
            return ResultTuple({}, {}, {})
        self.logger.info(
            "Will download %d products using %d workers",
            len(product_ids),
            self.n_concurrent_dl,
        )

        (
            statuses,
            online_prods,
            offline_prods,
            product_infos,
            exceptions,
        ) = self._init_statuses(product_ids)

        # Skip already downloaded files.
        # Although the download method also checks, we do not need to retrieve such
        # products from the LTA and use up our quota.
        if not overwrite:  # @evanjt skip this check if overwrite is True
            self._skip_existing_products(
                directory, offline_prods, product_infos, statuses, exceptions
            )

        stop_event = threading.Event()
        dl_tasks = {}
        trigger_tasks = {}

        # Two separate threadpools for downloading and triggering of retrieval.
        # Otherwise triggering might take up all threads and nothing is downloaded.
        dl_count = len(online_prods) + len(offline_prods)
        dl_executor = ThreadPoolExecutor(
            max_workers=max(1, min(self.n_concurrent_dl, dl_count)),
            thread_name_prefix="dl",
        )
        dl_progress = self._tqdm(
            total=dl_count,
            desc="Downloading products",
            unit="product",
        )
        if offline_prods:
            trigger_executor = ThreadPoolExecutor(
                max_workers=min(
                    self.api.concurrent_lta_trigger_limit, len(offline_prods)
                ),
                thread_name_prefix="trigger",
            )
            trigger_progress = self._tqdm(
                total=len(offline_prods),
                desc="LTA retrieval",
                unit="product",
                leave=True,
            )
        try:
            # First all online products are downloaded. Subsequently, offline products that might
            # have become available in the meantime are requested.
            for pid in itertools.chain(online_prods, offline_prods):
                future = dl_executor.submit(
                    self._download_online_retry,
                    product_infos[pid],
                    directory,
                    statuses,
                    exceptions,
                    stop_event,
                )
                dl_tasks[future] = pid

            for pid in offline_prods:
                future = trigger_executor.submit(
                    self._trigger_and_wait,
                    pid,
                    stop_event,
                    statuses,
                )
                trigger_tasks[future] = pid

            for task in concurrent.futures.as_completed(
                list(trigger_tasks) + list(dl_tasks)
            ):
                pid = trigger_tasks.get(task) or dl_tasks[task]
                exception = exceptions.get(pid)
                if task.cancelled():
                    exception = concurrent.futures.CancelledError()
                if task.exception():
                    exception = task.exception()

                if task in dl_tasks:
                    if not exception:
                        product_infos[pid] = task.result()
                        statuses[pid] = DownloadStatus.DOWNLOADED
                    dl_progress.update()
                    # Keep the LTA progress fresh
                    if offline_prods:
                        trigger_progress.update(0)
                else:
                    trigger_progress.update()
                    if all(t.done() for t in trigger_tasks):
                        trigger_progress.close()

                if exception:
                    exceptions[pid] = exception
                    if self.fail_fast:
                        raise exception from None
                    else:
                        self.logger.error(
                            "%s failed: %s", pid, _format_exception(exception)
                        )
        except:
            stop_event.set()
            for t in list(trigger_tasks) + list(dl_tasks):
                t.cancel()
            raise
        finally:
            dl_executor.shutdown()
            dl_progress.close()
            if offline_prods:
                trigger_executor.shutdown()
                trigger_progress.close()

        if not any(statuses):
            if not exceptions:
                raise SentinelAPIError(
                    "Downloading all products failed for an unknown reason"
                )
            exception = list(exceptions)[0]
            raise exception

        # Update Online status in product_infos
        for pid, status in statuses.items():
            if status in [DownloadStatus.OFFLINE, DownloadStatus.TRIGGERED]:
                product_infos[pid]["Online"] = False
            elif status != DownloadStatus.UNAVAILABLE:
                product_infos[pid]["Online"] = True

        return ResultTuple(statuses, exceptions, product_infos)



from typing import Tuple, List, Dict, Any, Union
from earth_extractor import core
from earth_extractor.core.models import CommonSearchResult
import logging
from earth_extractor.satellites import enums
from earth_extractor.satellites.base import Satellite
from earth_extractor.cli_options import (
    Satellites,
    TemporalFrequency,
    SatelliteChoices,
)
import shapely
import shapely.ops
from shapely.geometry import GeometryCollection
import pyproj
import datetime
import pandas as pd
from dataclasses import asdict
import requests
import os
import tqdm
import tenacity
from concurrent.futures import ThreadPoolExecutor, as_completed


# Define logger for this module
logger = logging.getLogger(__name__)
logger.setLevel(core.config.constants.LOGLEVEL_MODULE_DEFAULT)


def pair_satellite_with_level(
    choice: SatelliteChoices,
) -> Tuple[Satellite, enums.ProcessingLevel]:
    """Splits the input satellite and level choice from the command line

        Expects a `<Satellite>:<ProcessingLevel>` string.

    Parameters
    ----------
    choice : str
        The satellite and level choice from the command line

    Returns
    -------
    Tuple[Satellite, ProcessingLevel]
        A tuple containing the satellite and level choice
    """

    satellite, processing_level = choice.split(":")

    # Check if the satellite is valid, if true assign class to a variable
    if satellite not in Satellites.__members__:
        raise ValueError(
            f"Invalid satellite choice: {satellite}. "
            f"Valid choices are: {Satellites.__members__}"
        )
    else:
        satellite_choice: Satellite = Satellites[satellite].value

    # Check if the processing level is valid, if true assign enum to a variable
    if processing_level not in enums.ProcessingLevel.__members__:
        raise ValueError(
            f"Invalid processing level choice: {processing_level}. "
            f"Valid choices are: {enums.ProcessingLevel.__members__}"
        )
    else:
        level_choice: enums.ProcessingLevel = enums.ProcessingLevel[
            processing_level
        ]

    # Validate that the satellite contains the chosen level
    if level_choice not in satellite_choice.processing_levels:
        raise ValueError(
            f"The chosen processing level {level_choice} "
            f"is not available for the chosen satellite {satellite_choice}."
        )

    return (satellite_choice, level_choice)


def roi_from_geojson(geojson: str) -> GeometryCollection:
    """Extracts the ROI from a GeoJSON file

    Parameters
    ----------
    geojson : str
        The path to the GeoJSON file

    Returns
    -------
    str
        The boundary as a string
    """

    with open(geojson, "r") as in_file:
        geojson_data = in_file.read()
        geometry = shapely.from_geojson(geojson_data)

    if isinstance(geometry, shapely.GeometryCollection):
        geometry = geometry.geoms[0]
        logger.warning(
            "GeoJSON is a collection, only considering first element: "
            f"{geometry}"
        )
    if geometry is None:
        raise ValueError(
            f"Could not extract geometry from GeoJSON file: {geojson}"
        )

    return geometry


def is_float(string):
    """Checks if a string is a float"""
    try:
        float(string)
        return True
    except ValueError:
        return False


def buffer_in_metres(
    input_geom: GeometryCollection,
    buffer_metres: Union[float, int],
    crs_input: int = 4326,
    crs_output: int = 4326,
    crs_buffer: int = 3857,
) -> GeometryCollection:
    """Adds a buffer to geometry in metres

    Parameters
    ----------
    input_geom : GeometryCollection
        The geometry to buffer
    buffer_metres : Union[float, int}
        The buffer distance in metres
    crs_input : int, optional
        The CRS of the input geometry, by default 4326 (EPSG: WGS84)
    crs_output : int, optional
        The CRS of the output geometry, by default 4326 (EPSG: WGS84)
    crs_buffer : int, optional
        The CRS to execute generation of the buffer distance, by
        default 3857 (EPSG: Web Mercator), uses the same datum (WGS84) and
        units in metres

    Returns
    -------
    GeometryCollection
        The buffered geometry
    """

    logger.info(f"Applying buffer of {buffer_metres} metres to ROI")
    # Define pyproj objects for the input, buffer and output CRS
    crs_in = pyproj.CRS.from_epsg(crs_input)
    crs_conv = pyproj.CRS.from_epsg(crs_buffer)
    crs_out = pyproj.CRS.from_epsg(crs_output)

    # Perform the transformation for input
    transformer = pyproj.Transformer.from_crs(crs_in, crs_conv, always_xy=True)
    projected_geom = shapely.ops.transform(transformer.transform, input_geom)

    # Perform the buffer in the buffer CRS
    buffered_geom = projected_geom.buffer(buffer_metres)

    # Perform the transformation for output
    transformer = pyproj.Transformer.from_crs(
        crs_conv, crs_out, always_xy=True
    )
    reprojected_buffered_geom = shapely.ops.transform(
        transformer.transform, buffered_geom
    )

    logger.debug(f"Adjusted ROI (post-buffer): {reprojected_buffered_geom}")

    return reprojected_buffered_geom


def parse_roi(
    roi: str,
    buffer: float,
) -> GeometryCollection:
    """Parses the ROI input from the command line"""

    if "<" in roi or ">" in roi:
        # In the case the user explicitly inputs < or > as per the help msg
        raise ValueError("Do not include the '<' or '>' in the ROI.")

    # Check if the input is a BBox, Point or a path to a GeoJSON file
    if (len(roi.split(",")) == 4) and all(  # BBox
        [core.utils.is_float(i) for i in roi.split(",")]
    ):
        # If str splits into 4 float compatible values, consider it a BBox
        roi_obj = core.models.BBox.from_string(roi).to_shapely()
    elif (len(roi.split(",")) == 2) and all(  # Point
        [core.utils.is_float(i) for i in roi.split(",")]
    ):
        if buffer == 0:
            raise ValueError("Buffer must be greater than 0 for a point ROI.")
        roi_obj = core.models.Point.from_string(roi).to_shapely()
    else:
        # Otherwise, consider it a path to a GeoJSON file
        try:
            roi_obj = roi_from_geojson(roi)
        except FileNotFoundError as e:
            raise FileNotFoundError(
                f"GeoJSON file not found: {roi}. If you are trying to "
                f"specify a BBox or Point, please ensure you are using the "
                f"correct format. See the help message for more information."
            ) from e
        if roi_obj.geom_type == "Point" and buffer == 0:
            raise ValueError("Buffer must be greater than 0 for a point ROI.")

    logger.debug(f"Input ROI: {roi_obj}")

    # Apply buffer to ROI if required
    if buffer > 0:
        roi_obj = buffer_in_metres(roi_obj, buffer)

    return roi_obj


def download_by_frequency(
    start_date: datetime.datetime,
    end_date: datetime.datetime,
    frequency: TemporalFrequency,
    query_results: List[core.models.CommonSearchResult],
    filter_field: str = "cloud_cover_percentage",
) -> List[core.models.CommonSearchResult]:
    """With the given start and end dates and a frequency, choose the
    satellite image with the best cloud_cover percentage and download it.
    If there are multiple images with the same cloud_cover percentage,
    choose the one with the least cloud_cover percentage.

    If there are multiple minimum values (for example if there are two or
    more images with the same cloud_cover percentage of 0), the latest
    image in the frequency period is chosen (this behaviour is defined by
    the pandas Grouper function).

    Parameters
    ----------
    start_date : datetime.datetime
        The start date of the search
    end_date : datetime.datetime
        The end date of the search
    frequency : TemporalFrequency
        The frequency of the search according to the enum
    query_results : List[core.models.CommonSearchResult]
        The query results, in the internal common format
    filter_field : str, optional
        The field to filter by, by default 'cloud_cover_percentage'

    Returns
    -------
    List[core.models.CommonSearchResult]
        The filtered query results
    """

    # Convert the query results to a list of dicts
    query_results_dict = [asdict(x) for x in query_results]

    # If filter_field is not in the query results, raise an error
    cleaned_query_results: List[Dict[Any, Any]] = []
    for result in query_results_dict:
        if result[filter_field] is None:
            logger.warning(
                f"Filter field {filter_field} not in query result "
                f"for ID: {result['product_id']} ({result['satellite']}), "
                "ignoring."
            )
            logger.debug(f"Query result ignored: {result}")
        else:
            cleaned_query_results.append(result)

    # Raise exception if there are no results
    if len(cleaned_query_results) == 0:
        raise ValueError(
            f"Filter field '{filter_field}', used to define this frequency "
            "filter is not in any of the query results, therefore, "
            "no results are returned and no download will be attempted. "
        )

    df = pd.DataFrame.from_dict(cleaned_query_results)
    # Set the 'time' column as the DataFrame index
    df.set_index("time", inplace=True)
    df.index = pd.to_datetime(df.index)

    date_range = pd.date_range(start_date, end_date)
    mask = (df.index > date_range[0]) & (df.index <= date_range[-1])
    df = df.loc[mask]

    # Find the minimum row for each grouping (Using name of the enum
    # supplied by the enum: D, M, W, Y), and joining back to the original df.
    # Then ilter df by frequency minimum for the 'cloud_cover_percentage' value
    filtered = df.merge(
        df.groupby(pd.Grouper(freq=frequency.name))["cloud_cover_percentage"]
        .min()
        .reset_index(),
        how="inner",
    )

    results: List[CommonSearchResult] = [
        core.models.CommonSearchResult(**x)
        for x in filtered.to_dict("records")
    ]

    for result in results:
        logger.debug(
            f"Interval filter results: {result.satellite} "
            f"({result.processing_level}), File: {result.filename}, "
            f"Time: {result.time}, "
            f"{filter_field}: {getattr(result, filter_field)}"
        )

    logger.info(
        f"Interval operation filtered {len(query_results_dict)} to "
        f"{len(results)} results on {results[0].satellite} "
        f"({results[0].processing_level})."
    )

    return results


@tenacity.retry(
    stop=tenacity.stop_after_attempt(
        core.config.constants.MAX_DOWNLOAD_ATTEMPTS
    ),
    wait=tenacity.wait_exponential(multiplier=1, min=4, max=10),
    reraise=True,
)
def download_with_progress(
    url: str,
    output_folder: str,
    headers: Dict[str, str] = {},
    overwrite: bool = False,
) -> None:
    """Downloads a file with a progress bar

    The output file is the same filename as the original, and is saved to
    the output_folder.

    This function is used for downloading binary files from data providers. If
    a HTML content-type is provided, it is assumed that the download failed
    and the function will raise an error. This is necessary to check for
    example in the case of NASA where some errors are returned as a 200: OK but
    with a HTML error message.

    Parameters
    ----------
    url : str
        The URL to download
    output_folder : str
        The output file base directory
    headers : Dict[str, str], optional
        The headers to use for the download, by default {}. This is necessary
        for some providers that require authentication.
    overwrite : bool, optional
        Whether to overwrite existing files, by default False
    """

    with requests.get(url, stream=True, headers=headers) as resp:
        resp.raise_for_status()
        total_size = int(resp.headers.get("content-length", -1))

        if "text/html" in resp.headers["Content-Type"]:
            # We definitely don't want to download HTML when
            # expecting a binary file. This is likely due to
            # an expired token, auth error or NASA issue.
            if (  # NASA Error
                "We are currently having issues verifying "
                "your request. Please try again later"
                in resp.content.decode("utf-8")
            ):
                error_msg = (
                    "NASA CMR: Server error in verifying "
                    f"request with URL '{url}'. This is most likely a server "
                    "error. Check the NASA Alerts & Issues page at "
                    "https://ladsweb.modaps.eosdis.nasa.gov/alerts-and-issues/"
                )
            else:
                error_msg = (
                    f"Server error in verifying request with URL '{url}'. "
                    f"This is most likely a server error. "
                )
            raise RuntimeError(error_msg)
        output_file = os.path.join(output_folder, url.split("/")[-1])

        # Check if the file already exists, then apply overwrite policy
        if os.path.exists(output_file):
            if overwrite is False:
                # Check length of existing file matches expected size
                local_filesize = os.path.getsize(output_file)
                if total_size != -1 and total_size != local_filesize:
                    logger.info(
                        f"File exists {output_file} but does not match "
                        f"server size (local: {local_filesize} remote: "
                        f"{total_size}). Redownloading file."
                    )
                else:
                    logger.warning(
                        f"File already exists: {output_file}, skipping "
                        "download."
                    )
                    return
            else:
                logger.warning(
                    f"File already exists: {output_file}, "
                    "overwriting existing file."
                )

        with open(output_file, "wb") as dest:
            with tqdm.tqdm(
                total=total_size,
                desc=url,
                unit="iB",
                unit_scale=True,
                unit_divisor=1024,
            ) as bar:
                for chunk in resp.iter_content(chunk_size=4096):
                    if chunk:  # filter out keep-alive new chunks
                        size = dest.write(chunk)
                        bar.update(size)

    # Check size of downloaded file
    output_size = os.path.getsize(output_file)
    if output_size == 0:
        raise ValueError(
            f"Downloaded file {output_file} is empty, retrying download "
            f"(max {core.config.constants.MAX_DOWNLOAD_ATTEMPTS} attempts)."
        )

    if total_size != -1 and total_size != os.path.getsize(output_file):
        raise ValueError(
            f"Downloaded file {output_file} is not the expected size, "
            f"retrying download"
        )
    logger.debug(f"Downloaded file: {output_file}, size: {output_size}")

    return


def download_parallel(
    urls: List[str],
    output_folder: str,
    headers: Dict[str, str] = {},
    overwrite: bool = False,
    processes: int = core.constants.DEFAULT_DOWNLOAD_THREADS,
) -> None:
    """Downloads the given URLs in parallel

    Giving a total progress bar for all downloads, and overall if possible.

    Parameters
    ----------
    urls : List[str]
        A list of URLs to download
    output_folder : str
        The output file base directory
    headers : Dict[str, str], optional
        The headers to use for the download, by default {}. This is necessary
        for some providers that require authentication.
    overwrite : bool, optional
        Whether to overwrite existing files, by default False
    processes : int, optional
        The number of processes to use for downloading, by default
        core.constants.DEFAULT_DOWNLOAD_THREADS
    """

    with ThreadPoolExecutor(max_workers=processes) as executor:
        # Map download_item function to the URLs in the specified column
        futures = {
            executor.submit(
                download_with_progress, url, output_folder, headers, overwrite
            ): url
            for url in urls
        }

        # Retrieve the results as they become available
        for future in as_completed(futures):
            url = futures[future]
            try:
                result = future.result()
                logger.debug(
                    f"Downloaded file successfully (threading): {url} "
                    f"({result})"
                )
            except Exception as e:
                logger.error(f"{url} generated an exception: {e}")


def download_all_satellites_in_parallel(
    query_results: List[Tuple[Satellite, List[CommonSearchResult]]],
    output_folder: str,
    overwrite: bool = False,
) -> None:
    """Executes each provider download function in parallel

    As each provider has a different method, this function does not attempt
    to consolidate one big list, but issue the `download_many()` into
    many threads so they can all run at the same time.

    It is up to this `download_many()` function to provide parallel downloads
    for that individual provider.

    Parameters
    ----------
    query_results : List[Tuple[Satellite, List[CommonSearchResult]]]
        The query results, in the internal common format
    output_folder : str
        The output file base directory
    overwrite : bool, optional
        Whether to overwrite existing files, by default False
    """

    with ThreadPoolExecutor(
        max_workers=core.constants.DEFAULT_DOWNLOAD_THREADS
    ) as executor:
        # Map download_item function to the URLs in the specified column
        # satellite, results = query_results
        futures = {
            executor.submit(
                satgroup[0].download_many,
                search_results=satgroup[1],
                download_dir=output_folder,
                overwrite=overwrite,
            ): satgroup
            for satgroup in query_results
        }

        # Retrieve the results as they become available
        for future in as_completed(futures):
            url = futures[future]
            try:
                result = future.result()
                logger.debug(
                    f"Downloaded file successfully (threading): {url} "
                    f"({result})"
                )
            except Exception as e:
                logger.error(f"{url} generated an exception: {e}")


def download(query_results, output_dir, parallel=False, overwrite=False):
    """The main download function called by the app

    This consolidates the parallel, overwrite functionality and logging that
    became large enough for the routines that exist in 3+ different CLI
    commands.

    Parameters
    ----------
    query_results : List[Tuple[Satellite, List[CommonSearchResult]]]
        The query results, in the internal common format
    output_dir : str
        The output file base directory
    parallel : bool, optional
        Whether to download in parallel, by default False
    overwrite : bool, optional
        Whether to overwrite existing files, by default False

    """

    if parallel is True:
        # Download all of the satellites in parallel at the same time
        download_all_satellites_in_parallel(
            query_results,
            output_dir,
            overwrite=overwrite,
        )
    else:
        # Download the results using the satellite's download provider
        for sat, res in query_results:
            if len(res) > 0:
                logger.info(
                    f"Downloading results for {sat}..." f"({len(res)} items)"
                )

                # Download the results
                sat.download_many(
                    search_results=res,
                    download_dir=output_dir,
                    overwrite=overwrite,
                )
    logger.info(
        f"Download complete. Your files are in {os.path.abspath(output_dir)}"
    )



from earth_extractor.satellites.base import Satellite
from earth_extractor.satellites import enums
from earth_extractor.providers import swiss_topo
from earth_extractor import core
import logging

logger = logging.getLogger(__name__)
logger.setLevel(core.config.constants.LOGLEVEL_MODULE_DEFAULT)


swissimage: Satellite = Satellite(
    query_provider=swiss_topo,
    download_provider=swiss_topo,
    name=enums.Satellite.SWISSIMAGE,
    description="SwissImage 10cm",
    processing_levels=[
        enums.ProcessingLevel.CM10,
        enums.ProcessingLevel.CM200,
    ],
    sensors=[],
    filters=[],
)



from typing import List
from earth_extractor.satellites import enums
from earth_extractor.providers.base import Provider
import functools
from earth_extractor import core
import logging

logger = logging.getLogger(__name__)
logger.setLevel(core.config.constants.LOGLEVEL_MODULE_DEFAULT)


class Satellite:
    def __init__(
        self,
        query_provider: Provider,
        download_provider: Provider,
        name: enums.Satellite,
        description: str,
        processing_levels: List[enums.ProcessingLevel],
        sensors: List[enums.Sensor],
        filters: List[enums.Filters],
    ) -> None:
        """A satellite's data can be queried from one provider and downloaded
        from another, defining these allows the methods to be interchanged
        """

        self._query_provider = query_provider
        self._download_provider = download_provider

        # General information about the satellite
        self.name = name
        self.description = description
        self.processing_levels = processing_levels
        self.sensors = sensors
        self.filters = filters

        # Use the methods from the query and download providers
        self.query = functools.partial(
            self._query_provider.query, satellite=self
        )
        self.download_many = functools.partial(
            self._download_provider.download_many,
        )

    @property
    def has_cloud_cover(self) -> bool:
        """Whether the satellite has a cloud cover filter"""

        return enums.Filters.CLOUD_COVER in self.filters

    def _validate_satellite_provider_compatiblity(
        self,
    ) -> None:
        """Check that the given provider is compatible with the satellite"""

        if self not in self._download_provider.satellites:
            raise ValueError(
                f"Satellite {self.name} not supported by download provider: "
                f"'{self._download_provider.name}'"
            )
        if self not in self._query_provider.satellites:
            raise ValueError(
                f"Satellite {self.name} not supported by query provider: "
                f"'{self._query_provider.name}'"
            )

    def __str__(self):
        """Return the name of the satellite as a string.

        Allows for the satellite to be used as a dictionary key
        """

        return str(self.name)

    def __repr__(self):
        """Return the representable name of the satellite as a string."""

        return str(self.name)



# flake8: noqa
from .config import constants, Constants
from .logging import setup_file_logger, ErrorFlagHandler
from .models import BBox
from .utils import pair_satellite_with_level
from .query import batch_query
from .credentials import get_credentials



from earth_extractor.providers import Provider
from earth_extractor import core
from typing import Any, List, TYPE_CHECKING, Optional

# from sentinelhub.api.catalog import SentinelHubCatalog
# from sentinelhub.config import SHConfig
from sentinelhub.data_collections import DataCollection
import logging
import datetime
from earth_extractor.satellites import enums

if TYPE_CHECKING:
    from earth_extractor.satellites.base import Satellite

logger = logging.getLogger(__name__)
logger.setLevel(core.config.constants.LOGLEVEL_MODULE_DEFAULT)

credentials = core.credentials.get_credentials()


class SinergiseSentinelHub(Provider):
    def query(
        self,
        satellite: "Satellite",
        processing_level: enums.ProcessingLevel,
        roi: core.models.BBox,
        start_date: datetime.datetime,
        end_date: datetime.datetime,
        cloud_cover: Optional[int] = None,
    ) -> List[Any]:
        """Query the Sinergise Sentinel Hub for data"""

        # Check that the provider's credentials that are needed are set
        self._check_credentials_exist()

        logger.info("Querying Sinergise Sentinel Hub")
        if (
            credentials.SINERGISE_CLIENT_SECRET
            or credentials.SINERGISE_CLIENT_ID
        ) is None:
            raise ValueError(
                "Sinergise Sentinel Hub credentials not found. "
                "Please set SINERGISE_CLIENT_SECRET and SINERGISE_CLIENT_ID "
                "in the config file."
            )

        # config = SHConfig(
        #     sh_client_id=credentials.SINERGISE_CLIENT_ID,
        #     sh_client_secret=credentials.SINERGISE_CLIENT_SECRET,
        # )
        # catalog = SentinelHubCatalog(config=config)

        logger.info(
            f"Satellite: {satellite.name} ({satellite.name}) "
            f"{processing_level.value}"
        )

        product_type = self.products.get(
            (satellite.name, processing_level), None
        )
        if not product_type:
            raise ValueError(
                f"Processing level {processing_level.value} not supported "
                f"by Copernicus Open Access Hub for satellite "
                f"{satellite.name}. Available processing levels: "
                f"{self.products}"
            )

        raise NotImplementedError("Sinergise Sentinel Hub not implemented")


sinergise: SinergiseSentinelHub = SinergiseSentinelHub(
    name="Sinergise",
    description="Sinergise Sentinel Hub",
    uri="https://www.sinergise.com",
    products={
        (enums.Satellite.SENTINEL2, enums.ProcessingLevel.L1C): [
            DataCollection.SENTINEL2_L1C
        ],
        (enums.Satellite.SENTINEL2, enums.ProcessingLevel.L2A): [
            DataCollection.SENTINEL2_L2A
        ],
    },
    credentials_required=["SINERGISE_CLIENT_SECRET", "SINERGISE_CLIENT_ID"],
)



# flake8: noqa
from .base import Provider
from .copernicus import CopernicusOpenAccessHub, copernicus_scihub
from .alaskan_satellite_facility import asf
from .sinergise import sinergise
from .nasa import nasa_cmr
from .swisstopo import swiss_topo



#!/usr/bin/env python3
import datetime
import typer
from typing import Annotated, List
import logging
from earth_extractor import core, cli_options, __version__
import atexit
import os
import sys


# Define a console handler
console_handler = logging.StreamHandler()
console_handler.setLevel(core.config.constants.LOGLEVEL_CONSOLE)
console_handler.setFormatter(core.config.constants.LOGFORMAT_CONSOLE)
logging.getLogger().addHandler(console_handler)

# Define a file handler for ERRORs
error_handler = core.logging.ErrorFlagHandler()  # Watches for ERRORs
logging.getLogger().addHandler(error_handler)
atexit.register(error_handler.print_status)  # Print error status on exit

# Define logger for this module
logger = logging.getLogger(__name__)
logger.setLevel(core.config.constants.LOGLEVEL_MODULE_DEFAULT)


# Initialise the Typer class
app = typer.Typer(
    no_args_is_help=True,
    add_completion=False,
    pretty_exceptions_show_locals=False,
)


@app.command()
def import_geojson(
    filename: str = typer.Option(
        ..., help="Path to the GeoJSON file to import"
    ),
    output_dir: str = typer.Option(
        core.config.constants.DEFAULT_DOWNLOAD_DIR,
        help="Output directory for the downloaded files.",
    ),
    no_confirmation: bool = typer.Option(
        False,
        "--no-confirmation",
        help="Do not ask for confirmation before downloading",
    ),
    parallel: bool = typer.Option(
        True, help="Download all results in parallel"
    ),
    overwrite: bool = typer.Option(
        False,
        "--overwrite",
        help="Overwrite existing files in the output directory",
    ),
) -> None:
    """Import a GeoJSON file with metadata to the database

    The GeoJSON file must be in the same format as the one exported by the
    export command.
    """

    # Check if the file exists
    if not os.path.exists(filename):
        raise FileNotFoundError(f"File not found: {filename}")

    # Import the GeoJSON file
    query_results = core.query.import_query_results(geojson_file=filename)

    # Prompt user for confirmation before downloading
    if not no_confirmation:
        typer.confirm(
            "Do you want to download the results? (use --no-confirmation to "
            "skip this prompt)",
            abort=True,
        )

    core.utils.download(query_results, output_dir, parallel, overwrite)


@app.command()
def batch(
    start: Annotated[
        datetime.datetime,
        typer.Option("--start", help="Start datetime of the search."),
    ],
    end: Annotated[
        datetime.datetime,
        typer.Option("--end", help="End datetime of the search."),
    ],
    satellites: Annotated[
        List[cli_options.SatelliteChoices],
        typer.Option(
            "--satellite",
            help="Satellite to consider. To add multiple satellites, "
            "use the option multiple times.",
            case_sensitive=False,
        ),
    ],
    roi: Annotated[  # Multichoice types are not possible yet, take a str
        str,  # then parse option in function
        typer.Option(  # https://github.com/tiangolo/typer/issues/140
            "--roi",
            help="Region of interest to consider. "
            "Format: <lon_min,lat_min,lon_max,lat_max> for boundary, "
            "<lon,lat> for a point, or path to a "
            "GeoJSON file (eg. <bounds.json>). All inputs are assumed to be "
            "projected in WGS84 (EPSG: 4326), and all point geometries must "
            "also use the buffer option.",
        ),
    ],
    buffer: float = typer.Option(
        0.0, "--buffer", help="Buffer to apply to the ROI in metres."
    ),
    output_dir: str = typer.Option(
        core.config.constants.DEFAULT_DOWNLOAD_DIR,
        help="Output directory for the downloaded files.",
    ),
    cloud_cover: int = typer.Option(
        100, "--cloud-cover", help="Maximum allowed cloud cover percentage."
    ),
    no_confirmation: bool = typer.Option(
        False,
        "--no-confirmation",
        help="Do not ask for confirmation before downloading",
    ),
    export: cli_options.ExportMetadataOptions = typer.Option(
        cli_options.ExportMetadataOptions.DISABLED.value,
        help="Export query results. Output can be used for editing and "
        "re-importing with the import command. If PIPE is chosen, the "
        "logger will be disabled and the output will be printed to "
        "stdout.",
        case_sensitive=False,
    ),
    results_only: bool = typer.Option(
        False,
        help="Only export the results of the query, do not download. "
        "This is useful for exporting the results of a query to a "
        "GeoJSON file and exiting the program without user input.",
    ),
    parallel: bool = typer.Option(
        True, help="Download all results in parallel"
    ),
    overwrite: bool = typer.Option(
        False,
        "--overwrite",
        help="Overwrite existing files in the output directory",
    ),
) -> None:
    """Batch download of satellite data with minimal user input"""

    # Setup the file logger
    core.logging.setup_file_logger(output_dir)

    # Disable the console logger if the user wants to pipe the output
    if export == cli_options.ExportMetadataOptions.PIPE.value:
        logging.getLogger().removeHandler(console_handler)

    query_results = core.query.batch_query(
        start=start,
        end=end,
        satellites=satellites,
        roi=roi,
        buffer=buffer,
        output_dir=output_dir,
        cloud_cover=cloud_cover,
        export=export,
        results_only=results_only,
    )
    # Sum results from all query results
    total_qty = sum([len(res) for sat, res in query_results])

    logger.info(f"Total (all satellites) results qty: {total_qty}")

    if total_qty == 0:
        sys.exit("Nothing to download")

    # Prompt user for confirmation before downloading
    if not no_confirmation:
        typer.confirm(
            "Do you want to download the results? (use --no-confirmation to "
            "skip this prompt)",
            abort=True,
        )

    core.utils.download(query_results, output_dir, parallel, overwrite)


@app.command()
def batch_interval(
    start: Annotated[
        datetime.datetime,
        typer.Option("--start", help="Start datetime of the search."),
    ],
    end: Annotated[
        datetime.datetime,
        typer.Option("--end", help="End datetime of the search."),
    ],
    satellites: Annotated[
        List[cli_options.SatelliteChoices],
        typer.Option(
            "--satellite",
            help="Satellite to consider. To add multiple satellites, "
            "use the option multiple times.",
            case_sensitive=False,
        ),
    ],
    roi: Annotated[  # Multichoice types are not possible yet, take a str
        str,  # then parse option in function
        typer.Option(  # https://github.com/tiangolo/typer/issues/140
            "--roi",
            help="Region of interest to consider. "
            "Format: <lon_min,lat_min,lon_max,lat_max> for boundary, "
            "<lon,lat> for a point, or path to a "
            "GeoJSON file (eg. <bounds.json>). All inputs are assumed to be "
            "projected in WGS84 (EPSG: 4326), and all point geometries must "
            "also use the buffer option.",
        ),
    ],
    frequency: Annotated[
        cli_options.TemporalFrequency,
        typer.Option(
            help="Frequency to download data at. "
            "Options: yearly, monthly, weekly, daily",
            case_sensitive=False,
        ),
    ],
    buffer: float = typer.Option(
        0.0, "--buffer", help="Buffer to apply to the ROI in metres."
    ),
    output_dir: str = typer.Option(
        core.config.constants.DEFAULT_DOWNLOAD_DIR,
        help="Output directory for the downloaded files.",
    ),
    cloud_cover: int = typer.Option(
        100, "--cloud-cover", help="Maximum allowed cloud cover percentage."
    ),
    no_confirmation: bool = typer.Option(
        False,
        "--no-confirmation",
        help="Do not ask for confirmation before downloading",
    ),
    export: cli_options.ExportMetadataOptions = typer.Option(
        cli_options.ExportMetadataOptions.DISABLED.value,
        help="Export query results. Output can be used for editing and "
        "re-importing with the import command. If PIPE is chosen, the "
        "logger will be disabled and the output will be printed to "
        "stdout.",
        case_sensitive=False,
    ),
    results_only: bool = typer.Option(
        False, help="Only export the results of the query, do not download"
    ),
    parallel: bool = typer.Option(
        True, help="Download all results in parallel"
    ),
    overwrite: bool = typer.Option(
        False,
        "--overwrite",
        help="Overwrite existing files in the output directory",
    ),
) -> None:
    """Batch download with best cloud cover over a given time period

    Behaves similar to the batch command, but instead of downloading all
    results, it will download the result with the best cloud cover for each
    satellite:level combination over the given time period.

    For example, if the user specifies a time period of 2020-01-01 to
    2020-01-31, and the user specifies Sentinel-2 L1C and L2A, the command will

    """
    # Setup the file logger
    core.logging.setup_file_logger(output_dir)

    # Disable the console logger if the user wants to pipe the output
    if export == cli_options.ExportMetadataOptions.PIPE.value:
        logging.getLogger().removeHandler(console_handler)

    query_results = core.query.batch_query(
        start=start,
        end=end,
        satellites=satellites,
        roi=roi,
        buffer=buffer,
        output_dir=output_dir,
        cloud_cover=cloud_cover,
        export=export,
        results_only=results_only,
        interval_frequency=frequency,
    )
    # Sum results from all query results
    total_qty = sum([len(res) for sat, res in query_results])

    logger.info(f"Total (all satellites) results qty: {total_qty}")

    if total_qty == 0:
        sys.exit("Nothing to download")

    # Prompt user for confirmation before downloading
    if not no_confirmation:
        typer.confirm(
            "Do you want to download the results? (use --no-confirmation to "
            "skip this prompt)",
            abort=True,
        )

    core.utils.download(query_results, output_dir, parallel, overwrite)


@app.command()
def credentials(
    delete: str = typer.Option(None, help="Value to delete"),
    set: bool = typer.Option(False, help="Set all credentials"),
    show_secrets: bool = typer.Option(
        False, help="Display values of saved secrets"
    ),
) -> None:
    """Management of service credential keys"""

    if set:
        logger.info(
            "Setting credentials. Press enter to accept default "
            "value given in brackets."
        )
        core.credentials.set_all_credentials()

    if delete:
        core.credentials.delete_credential(delete)

    core.credentials.show_credential_list(show_secret=show_secrets)


def version_cb(value: bool) -> None:
    """Prints the version number of earth-extractor

    Parameters
    ----------
    value : bool
        The value of the --version option
    """

    if value:  # Only run on when --version is set
        typer.echo(f"earth-extractor {__version__}")
        sys.exit()


@app.callback()
def menu(
    version: bool = typer.Option(
        False,
        "--version",
        help="Prints the version number of earth-extractor",
        callback=version_cb,
        is_eager=True,
    ),
):
    """EarthExtractor - A tool for downloading satellite data from various
    providers. It is designed to be simple to use, and to provide a consistent
    interface for downloading data from different providers.

    Use the --help option on a subcommand to see more information about it.
    """
    pass


def main() -> None:
    """The main function of the application

    Used by the poetry entrypoint.
    """

    app()


if __name__ == "__main__":
    main()



from earth_extractor.providers import Provider
import logging
from typing import List
import asf_search
from earth_extractor.core.credentials import get_credentials
from earth_extractor import core
from earth_extractor.providers.extensions.asf_search import granule_search

logger = logging.getLogger(__name__)
logger.setLevel(core.config.constants.LOGLEVEL_MODULE_DEFAULT)

credentials = get_credentials()


class AlaskanSateliteFacility(Provider):
    def download_many(
        self,
        search_results: List[core.models.CommonSearchResult],
        download_dir: str,
        overwrite: bool = False,
        processes: int = 6,
    ) -> None:
        """Download many files from the Alaskan Satellite Facility"""

        # Check that the provider's credentials that are needed are set
        self._check_credentials_exist()

        # Extract the file ids from the search results
        search_file_ids = [
            result.identifier
            for result in search_results
            if result.identifier is not None
        ]

        logger.info("Downloading from Alaskan Satellite Facility")
        logger.debug(f"Search file ids: {search_file_ids}")

        # Add the ASF logger to the root logger
        asf_logger = logging.getLogger("asf_search")
        asf_logger.setLevel(core.config.constants.LOGLEVEL_MODULE_DEFAULT)

        self.create_download_folder(download_dir)  # Create the download folder

        # Authenticate with ASF
        try:
            session = asf_search.ASFSession().auth_with_token(
                token=credentials.NASA_TOKEN,
            )

            # Search for the granules
            res = granule_search(search_file_ids)
            logger.info(
                f"Found {len(res)} files to download (may include "
                "metadata files)"
            )

            if len(res) == 0:
                logger.info("No files to download, skipping")
                return  # nothing to download

            # Download the granules using the ASF API library
            res.download(
                path=download_dir,
                session=session,
                processes=processes,
                overwrite=overwrite,
            )
        except asf_search.ASFAuthenticationError as e:
            logger.error(
                "ASF authentication error. Check your credentials and make "
                "sure that you have accepted the ASF EULA in your NASA "
                "profile at "
                "https://urs.earthdata.nasa.gov/users/ejayt/unaccepted_eulas"
            )
            logger.debug(f"ASF Authentication error: {e}")
            return


asf: AlaskanSateliteFacility = AlaskanSateliteFacility(
    name="Alaskan Satellite Facility",
    description="Alaskan Satellite Facility",
    uri="https://asf.alaska.edu",
    credentials_required=["NASA_TOKEN"],
)



from enum import Enum
from earth_extractor.satellites import sentinel, modis, viirs, swissimage


class SatelliteChoices(str, Enum):
    # Sentinel
    SENTINEL1L1 = "SENTINEL1:L1"
    SENTINEL2L1C = "SENTINEL2:L1C"
    SENTINEL2L2A = "SENTINEL2:L2A"
    SENTINEL3L1 = "SENTINEL3:L1"
    SENTINEL3L2 = "SENTINEL3:L2"

    # MODIS
    MODISTERRA = "MODIS_TERRA:L1B"
    MODISAQUA = "MODIS_AQUA:L1B"

    # VIIRS
    VIIRS = "VIIRS:L1"

    # SwissImage
    SWISSIMAGE10CM = "SWISSIMAGE:CM10"
    SWISSIMAGE200CM = "SWISSIMAGE:CM200"


class Satellites(Enum):
    SENTINEL1 = sentinel.sentinel_1
    SENTINEL2 = sentinel.sentinel_2
    SENTINEL3 = sentinel.sentinel_3
    MODIS_TERRA = modis.modis_terra
    MODIS_AQUA = modis.modis_aqua
    VIIRS = viirs.viirs
    SWISSIMAGE = swissimage.swissimage


class TemporalFrequency(str, Enum):
    """Enum for temporal frequency to be used for

    In reverse as Typer only uses values and not names for selection in CLI
    """

    D = "DAILY"
    W = "WEEKLY"
    M = "MONTHLY"
    Y = "YEARLY"


class ExportMetadataOptions(str, Enum):
    DISABLED = "DISABLED"
    FILE = "FILE"
    PIPE = "PIPE"



from earth_extractor.satellites.base import Satellite
from earth_extractor.satellites import enums
from earth_extractor.providers import nasa_cmr
from earth_extractor import core
import logging

logger = logging.getLogger(__name__)
logger.setLevel(core.config.constants.LOGLEVEL_MODULE_DEFAULT)

""" Define the VIIRS constellation """

viirs: Satellite = Satellite(
    query_provider=nasa_cmr,
    download_provider=nasa_cmr,
    name=enums.Satellite.VIIRS,
    description="Visible Infrared Imaging Radiometer Suite",
    processing_levels=[enums.ProcessingLevel.L1],
    sensors=[enums.Sensor.VIIRS],
    filters=[],
)



from earth_extractor.satellites.base import Satellite
from earth_extractor.satellites import enums
from earth_extractor.providers import copernicus_scihub, asf
from earth_extractor import core
import logging

logger = logging.getLogger(__name__)
logger.setLevel(core.config.constants.LOGLEVEL_MODULE_DEFAULT)

""" Define the Sentinel satellites

    The satellites are defined as instances of the Satellite class.
"""

sentinel_1: Satellite = Satellite(
    query_provider=copernicus_scihub,
    download_provider=asf,
    name=enums.Satellite.SENTINEL1,
    description="Sentinel 1",
    processing_levels=[enums.ProcessingLevel.L1, enums.ProcessingLevel.L2],
    sensors=[enums.Sensor.C_SAR],
    filters=[],
)

sentinel_2: Satellite = Satellite(
    query_provider=copernicus_scihub,
    download_provider=copernicus_scihub,
    name=enums.Satellite.SENTINEL2,
    description="Sentinel 2",
    processing_levels=[enums.ProcessingLevel.L1C, enums.ProcessingLevel.L2A],
    sensors=[enums.Sensor.MSI],
    filters=[enums.Filters.CLOUD_COVER],
)

sentinel_3: Satellite = Satellite(
    query_provider=copernicus_scihub,
    download_provider=copernicus_scihub,
    name=enums.Satellite.SENTINEL3,
    description="Sentinel 3",
    processing_levels=[enums.ProcessingLevel.L1, enums.ProcessingLevel.L2],
    sensors=[enums.Sensor.OLCI, enums.Sensor.SLSTR, enums.Sensor.SRAL],
    filters=[enums.Filters.CLOUD_COVER],
)



from typing import Any, Dict, List, Tuple, TYPE_CHECKING, Optional
from earth_extractor import core
import logging
import datetime
import os
import shapely.geometry
from earth_extractor.core.models import CommonSearchResult
from earth_extractor.core.credentials import get_credentials
from pystac_client import Client

if TYPE_CHECKING:
    from earth_extractor.satellites import enums
    from earth_extractor.satellites.base import Satellite

logger = logging.getLogger(__name__)
logger.setLevel(core.config.constants.LOGLEVEL_MODULE_DEFAULT)

credentials = get_credentials()


class Provider:
    def __init__(
        self,
        name: str,
        description: Optional[str] = None,
        uri: Optional[str] = None,
        products: Dict[
            Tuple["enums.Satellite", "enums.ProcessingLevel"], List[Any]
        ] = {},
        credentials_required: List[Optional[str]] = [],
    ):
        """A provider of satellite data

        Parameters
        ----------
        name : str
            The name of the provider
        description : str, optional
            A description of the provider, by default None
        uri : str, optional
            A URI for the provider, by default None
        products : Dict[Tuple[Satellite, ProcessingLevel], List[Any]], optional
            A dictionary of the products supported by the provider. The key
            is a tuple of the satellite enum and the processing level enum.
            The value is a list of the products supported by the provider and
            in most cases is a single item, however, some providers may provide
            multiple products for a given satellite and processing level. For
            the sake of simplicity, the only case that does this at the time
            of this writing is the Copernicus Open Access Hub, which provides
            LFR and WFR products for Sentinel3 L2 data.
            )
            For example:
            {
                (enums.Satellite.SENTINEL1, enums.ProcessingLevel.GRD):
                    ["IW", "EW"]
            }
            by default {}
        credentials_required : List[Optional[str]], optional
            A list of credentials required by the provider, by default []. The
            value should be a string representation of the field in the
            earth_extractor.core.credentials.Credentials class. For example,
            the Copernicus Open Access Hub requires the credentials
            SCIHUB_USERNAME and SCIHUB_PASSWORD, so the value would be
            ["SCIHUB_USERNAME", "SCIHUB_PASSWORD"]. A check can be done on the
            existance of these vars using the _check_credentials_exist() method
            to determine whether these credentials are set, and if not, an
            error is raised.
        Raises
        ------
        NotImplementedError
            If the query method is not implemented
        """
        self.name = name
        self.description = description
        self.uri = uri
        self.products = products
        self.credentials_required = credentials_required

    def query(
        self,
        satellite: "Satellite",
        processing_level: "enums.ProcessingLevel",
        roi: core.models.BBox,
        start_date: datetime.datetime,
        end_date: datetime.datetime,
        cloud_cover: int = 100,
    ) -> List[CommonSearchResult]:
        """Query the provider for items matching the given parameters"""

        raise NotImplementedError(
            f"Query method not implemented for Provider: {self.name} "
            f"({self.description})"
        )

    def download_many(
        self,
        search_results: List[CommonSearchResult],
        download_dir: str,
        overwrite: bool = False,
        processes: int = 6,
    ) -> Any:
        raise NotImplementedError(
            f"Download method not implemented for Provider: {self.name} "
            f"({self.description})"
        )

    def create_download_folder(
        self,
        folder_name: str,
    ) -> None:
        """Create a download folder if it doesn't exist"""

        if not os.path.exists(folder_name):
            os.makedirs(folder_name)

    def translate_search_results(
        self,
        *,
        provider_search_results: Dict[Any, Any],
    ) -> List[core.models.CommonSearchResult]:
        raise NotImplementedError(
            "Search translation method not implemented for Provider: "
            f"{self.name} ({self.description})"
        )

    @property
    def _products_reversed(
        self,
    ) -> Dict[Any, Tuple["enums.Satellite", "enums.ProcessingLevel"]]:
        """A reverse of the property dictionary

        Makes sure that each item in the list each value is now its own key
        and the original key is now a value in the list

        For example, a dictionary of the form:
        {
            ("Sentinel-1", "GRD"): ["IW", "EW"]
        }
        Will become a dictionary of the form:
        {
            "IW": ("Sentinel-1", "GRD"),
            "EW": ("Sentinel-1", "GRD")
        }
        """

        products_reversed = {}
        for key, value in self.products.items():
            for item in value:
                products_reversed[item] = key

        return products_reversed

    def query_stac(
        self,
        provider_uri: str,
        collections: List[str],
        roi: shapely.geometry.base.BaseGeometry,
        start_date: datetime.datetime,
        end_date: datetime.datetime,
    ) -> Dict[str, str]:
        """A generic STAC query method for providers that support STAC

        Input geometry will be converted to a bounding box as some queries will
        fail with complex geometries.

        Parameters
        ----------
        provider_uri : str
            The URI of the provider
        collections : List[str]
            A list of collections to query
        roi : shapely.geometry.base.BaseGeometry
            The region of interest
        start_time : datetime.datetime
            The start time of the query
        end_time : datetime.datetime
            The end time of the query

        Returns
        -------
        Dict[str, str]
            A dictionary of the results
        """

        catalog = Client.open(provider_uri)

        # Convert roi to bbox STAC does not like complicated geometries
        roi_bbox = shapely.geometry.box(*roi.bounds)

        logger.info(f"Querying STAC URI: {provider_uri}")
        search = catalog.search(
            collections=collections,
            intersects=roi_bbox,
            datetime=[start_date.isoformat(), end_date.isoformat()],
        )

        return search.item_collection_as_dict()

    def _check_credentials_exist(self) -> None:
        for credential in self.credentials_required:
            if getattr(credentials, credential) is None:
                raise ValueError(
                    f"Credential '{credential}' for {self.name} is required "
                    "but its value has not been set, please set the "
                    "credential with the command "
                    "`earth-extractor credentials --set`"
                )



from pydantic import BaseModel, Field, AnyUrl
import shapely.geometry
import shapely.wkt
from typing import Any, TYPE_CHECKING, Optional
from dataclasses import dataclass, asdict
import datetime
from enum import Enum


if TYPE_CHECKING:
    from earth_extractor.satellites.enums import (
        ProcessingLevel,
        Sensor,
        Satellite,
    )


class BBox(BaseModel):
    """Defines the region of interest to be used for the search

    The region of interest is defined by a list of floats. The first two
    floats define the top left corner of the rectangle, and the last two
    floats define the bottom right corner of the rectangle. The coordinates
    are in the WGS84 coordinate system.
    """

    # Limit latitude to -90 to 90
    latmin: float = Field(..., ge=-90, le=90)
    lonmin: float = Field(..., ge=-180, le=180)
    latmax: float = Field(..., ge=-90, le=90)
    lonmax: float = Field(..., ge=-180, le=180)

    @classmethod
    def from_string(cls, v: str):
        """Convert a string into a list of floats

        The string is split by a comma (',') and each element is converted
        into a float. The resulting list is returned.
        """

        if isinstance(v, str):
            values = [float(x) for x in v.split(",")]

            return cls(
                latmin=values[0],
                lonmin=values[1],
                latmax=values[2],
                lonmax=values[3],
            )

    def __str__(self):
        return f"{self.latmin},{self.lonmin},{self.latmax},{self.lonmax}"

    def to_shapely(self) -> shapely.geometry.box:
        """Convert the ROI into a shapely object

        The shapely object is a rectangle defined by the top left and bottom
        right corners of the ROI.
        """

        return shapely.geometry.box(
            self.lonmin, self.latmin, self.lonmax, self.latmax
        )


class Point(BaseModel):
    """Defines the point of interest to be used for the search"""

    # Limit latitude to -90 to 90
    lat: float = Field(..., ge=-90, le=90)
    lon: float = Field(..., ge=-180, le=180)

    @classmethod
    def from_string(cls, v: str):
        """Convert a string into a list of floats

        The string is split by a comma (',') and each element is converted
        into a float. The resulting list is returned.
        """

        if isinstance(v, str):
            values = [float(x) for x in v.split(",")]

            return cls(lat=values[1], lon=values[0])

    def __str__(self):
        return f"{self.lat},{self.lon}"

    def to_shapely(self):
        """Convert the ROI into a shapely object

        The shapely object is a rectangle defined by the top left and bottom
        right corners of the ROI.
        """

        return shapely.geometry.Point(self.lon, self.lat)


@dataclass
class CommonSearchResult:
    """A class to support the exchange of search results between providers

    This class is used to standardise the search results between providers
    such that the the Provider/Satellite classes can be agnostic to the
    provider of the search results.
    """

    # Necessary for importing geojson to be able to determine provider
    satellite: "Satellite"

    product_id: Optional[Any] = None
    link: Optional[AnyUrl] = None
    identifier: Optional[str] = None
    filename: Optional[str] = None

    time: Optional[datetime.datetime] = None

    cloud_cover_percentage: Optional[float] = None
    size: Optional[float] = None

    processing_level: Optional["ProcessingLevel"] = None
    sensor: Optional["Sensor"] = None

    geometry: Optional[str] = None  # Should be as WKT
    url: Optional[AnyUrl] = None
    notes: Optional[str] = None

    # Create function to convert this dataclass to geojson including all fields
    # and casting enums to their values
    def to_geojson(self):
        """Convert this dataclass to geojson including all fields and casting
        enums to their values"""

        # Convert the dataclass to a dictionary
        d = asdict(self)

        # Convert all enums to their values
        for k, v in d.items():
            if isinstance(v, Enum):
                d[k] = v.value
        import geojson

        geom = None
        # Convert WKT geometry to shapely object
        if self.geometry is not None and isinstance(self.geometry, str):
            geom = shapely.wkt.loads(self.geometry)

        # Convert datetime to string
        if self.time is not None:
            d["time"] = self.time.isoformat()

        d.pop("geometry")  # Don't include geometry in the properties

        # Convert the dictionary to a geojson
        return geojson.Feature(geometry=geom, **d)



#!/usr/bin/env python3
import datetime
from typing import List, Tuple, Dict, Optional
import logging
from earth_extractor import core, cli_options
import os
import sys
import orjson
from earth_extractor.satellites import enums
from earth_extractor.satellites.base import Satellite
from earth_extractor.core.models import CommonSearchResult
import geopandas as gpd
import pandas as pd


# Define logger for this module
logger = logging.getLogger(__name__)
logger.setLevel(core.config.constants.LOGLEVEL_MODULE_DEFAULT)


def construct_geojson(
    gdf: gpd.GeoDataFrame,
    satellites: List[str],
    roi: str,
    buffer: float,
    cloud_cover: int,
    start: datetime.datetime,
    end: datetime.datetime,
    interval_frequency: Optional[cli_options.TemporalFrequency] = None,
    output_file: Optional[str] = None,
) -> None:
    """Converts a geodataframe to geojson

    If output_dir is empty, the geojson is printed to stdout. If output_dir
    is not empty, the geojson is exported to a file.

    The additional parameters are added to the geojson as the original query
    parameters in the metadata section `query_parameters`.

    Parameters
    ----------
    gdf : gpd.GeoDataFrame
        The GeoDataFrame
    satellites : List[str]
        The satellites IDs as a list of strings
    roi : str
        The ROI
    buffer : float
        The buffer
    cloud_cover : int
        The cloud cover
    start : datetime.datetime
        The start date
    end : datetime.datetime
        The end date
    interval_frequency : Optional[cli_options.TemporalFrequency], optional
        The interval frequency, by default None
    output_file : Optional[str], optional
        The output filename, by default None. If None, the geojson is printed
        to stdout

    """

    geojson_data = orjson.loads(gdf.to_json())
    geojson_data["query_parameters"] = {
        "satellites": satellites,
        "roi": roi,
        "buffer": buffer,
        "cloud_cover": cloud_cover,
        "start_date": start.isoformat(),
        "end_date": end.isoformat(),
        "interval_frequency": (
            interval_frequency.value if interval_frequency else None
        ),
    }
    if output_file is None:
        # Print the geojson to stdout (PIPE option)
        print(orjson.dumps(geojson_data).decode("utf-8"))
    else:
        logger.info(f"Exporting results to GeoJSON: {output_file}")

        # Export the geojson to a file with an indent of 2 spaces
        with open(os.path.join(output_file), "wb") as f:
            f.write(orjson.dumps(geojson_data, option=orjson.OPT_INDENT_2))


def convert_query_results_to_geodataframe(
    query_results: List[CommonSearchResult],
) -> gpd.GeoDataFrame:
    """Converts the query results to a GeoDataFrame

    Parameters
    ----------
    query_results : List[CommonSearchResult]
        The query results

    Returns
    -------
    gpd.GeoDataFrame
        The query results as a GeoDataFrame
    """

    # Convert the query results to a list of dicts
    query_results_geojson = [x.to_geojson() for x in query_results]

    # Convert the query results to a GeoDataFrame
    gdf = gpd.GeoDataFrame.from_dict(query_results_geojson)

    return gdf


def convert_geodataframe_to_query_results(
    gdf: gpd.GeoDataFrame,
) -> List[CommonSearchResult]:
    """Converts the GeoDataFrame to query results

    A reverse of the `convert_query_results_to_geodataframe` function, used
    in importing the exported GeoJSON files.

    Parameters
    ----------
    gdf : gpd.GeoDataFrame
        The GeoDataFrame

    Returns
    -------
    List[CommonSearchResult]
        The query results
    """

    # Convert the GeoDataFrame to a list of dicts
    query_results = [x for x in gdf.to_dict("records")]
    query_results = []
    for idx, row in gdf.iterrows():
        query_results.append(
            CommonSearchResult(
                satellite=row["satellite"],
                product_id=row["product_id"],
                link=row["link"],
                identifier=row["identifier"],
                filename=row["filename"],
                time=row["time"],
                cloud_cover_percentage=row["cloud_cover_percentage"],
                size=row["size"],
                processing_level=row["processing_level"],
                sensor=row["sensor"],
                geometry=row["geometry"],
            )
        )

    return query_results


def batch_query(
    start: datetime.datetime,
    end: datetime.datetime,
    satellites: List[cli_options.SatelliteChoices],
    roi: str,
    buffer: float,
    output_dir: str,
    cloud_cover: int,
    export: cli_options.ExportMetadataOptions,
    results_only: bool,
    interval_frequency: Optional[cli_options.TemporalFrequency] = None,
) -> List[Tuple[Satellite, List[CommonSearchResult]]]:
    if results_only:
        if export == cli_options.ExportMetadataOptions.DISABLED.value:
            raise ValueError("Cannot use --results-only without --export")
        logger.info("Skipping download, only exporting results")

    if interval_frequency is None:
        logger.info("Starting Earth Extractor batch mode")
    else:
        logger.info(
            "Starting Earth Extractor interval-batch mode with "
            f"interval frequency {interval_frequency.value}"
        )

    logger.info(f"Time: {start} {end}")
    roi_obj = core.utils.parse_roi(roi, buffer)

    # Parse the satellite:level string into a list of workable tuples
    satellite_operations: List[Tuple[Satellite, enums.ProcessingLevel]] = []

    # Create a list of tuples of satellites and levels
    for sat in satellites:
        satellite_operations.append(core.utils.pair_satellite_with_level(sat))

    # Perform a query for each satellite and level and append it to a list
    all_results = []

    # Create an empty geodataframe to hold data that can be appended
    # with all the satellite results
    gdf_all = gpd.GeoDataFrame()

    # Process each satellite and level
    for sat, level in satellite_operations:
        logger.debug(
            f"Querying satellite Satellite: {sat}, Level: {level.value}"
        )

        if cloud_cover < 100 and not sat.has_cloud_cover:
            logger.warning(
                f"Satellite {sat} does not support cloud cover filtering. "
                "Ignoring the filter and continuing."
            )
        # If the satellite does not have a cloud_cover query parameter, set it
        # to None
        res = sat.query(
            processing_level=level,
            roi=roi_obj,
            start_date=start,
            end_date=end,
            cloud_cover=cloud_cover if sat.has_cloud_cover else None,
        )

        if interval_frequency is not None:
            # If interval frequency is set, then we need to query the results
            # by the frequency
            res = core.utils.download_by_frequency(
                start_date=start,
                end_date=end,
                frequency=interval_frequency,
                query_results=res,
                filter_field="cloud_cover_percentage",
            )

        msg_summary = (
            f"Satellite: {sat}, Level: {level.value}. "
            f"Results qty: ({len(res)})"
        )

        # If the user wants to export the results, do so
        if export != cli_options.ExportMetadataOptions.DISABLED.value:
            # Add the results to the geodataframe
            results = convert_query_results_to_geodataframe(res)
            if not results.empty:
                gdf_all = pd.concat([gdf_all, results])
            else:
                logger.info(
                    f"No results for satellite {sat} and level {level.value}"
                )
                continue

            if sat == cli_options.Satellites.SWISSIMAGE.value:
                # Warn the user that swiss image 10 and 200cm data is not
                # going to have the correct geometry. Do this only in the
                # FILE output as to not corrupt the PIPE output.
                logger.warning(
                    "Due to the limits of the SwissImage API, the "
                    "boundaries of SwissImage will reflect only the "
                    "geometry of the given ROI in the query. See comments "
                    "within the code for more information."
                )
        logger.info(msg_summary)  # As to not pollute PIPE output

        # Append results to a list with associated satellite in order to use
        # its defined download provider
        all_results.append((sat, res))

    if export != cli_options.ExportMetadataOptions.DISABLED.value:
        satellite_list = [sat.value for sat in satellites]
        output_file = None  # Default PIPE output

        if export == cli_options.ExportMetadataOptions.FILE.value:
            # Remove millisecondsfrom the timestamp
            timestamp = core.config.constants.COMMON_TIMESTAMP.split(".")[0]
            timestamp = timestamp.replace("-", "")  # Remove symbols
            output_file = os.path.join(
                # Common timestamp without milliseconds and with summary
                # of satellites
                # eg: 20230824T085346_swissimage-cm200_swissimage-cm10.geojson
                output_dir,
                f"{timestamp}_"
                f"{'_'.join(satellite_list).replace(':', '-').lower()}"
                f".geojson",
            )

        construct_geojson(
            gdf=gdf_all,
            satellites=satellite_list,
            roi=roi,
            buffer=buffer,
            cloud_cover=cloud_cover,
            start=start,
            end=end,
            interval_frequency=interval_frequency,
            output_file=output_file,
        )

    if results_only or export == cli_options.ExportMetadataOptions.PIPE.value:
        """Exit on results only or PIPE export

        If the application does not exit on PIPE, the user's command will
        not stop executing
        """
        sys.exit()

    return all_results


def import_query_results(
    geojson_file: str,
) -> List[Tuple[Satellite, List[CommonSearchResult]]]:
    """Import a geojson file, convert it CommonSearchResult objects"""
    logger.info("Starting Earth Extractor import mode")
    logger.info(f"Importing geojson file: {geojson_file}")

    # Read the geojson file
    import geopandas

    gdf = geopandas.read_file(geojson_file)

    # Convert the geojson file to a list of CommonSearchResult objects
    results = convert_geodataframe_to_query_results(gdf)

    # Create a tuple of satellite and grouping of its results
    satellite_groups: Dict[Satellite, List[CommonSearchResult]] = {}
    for result in results:
        satellite_choice = cli_options.Satellites[result.satellite].value

        # Add satellite to dict if not already present
        if satellite_choice not in satellite_groups:
            satellite_groups[satellite_choice] = []
        # Add result to satellite group list
        satellite_groups[satellite_choice].append(result)

    # Convert the dict to a list of tuples
    query_results = [(key, rows) for key, rows in satellite_groups.items()]

    total_qty = sum([len(res) for sat, res in query_results])
    logger.info(
        f"Imported {total_qty} records from {len(query_results)} "
        f"satellite{'(s)' if len(query_results) > 1 else ''}"
    )
    return query_results



from earth_extractor.providers import Provider
from earth_extractor.core.credentials import get_credentials
from earth_extractor.core.models import CommonSearchResult
from typing import Any, List, TYPE_CHECKING, Dict, Tuple, Optional
import sentinelsat
from earth_extractor.providers.extensions.sentinelsat import (
    SentinelAPIExtended,
)
import logging
import datetime
import shapely
import shapely.geometry
from earth_extractor.satellites import enums
from earth_extractor import core
import tenacity


if TYPE_CHECKING:
    from earth_extractor.satellites.base import Satellite

logger = logging.getLogger(__name__)
logger.setLevel(core.config.constants.LOGLEVEL_MODULE_DEFAULT)

credentials = get_credentials()


class CopernicusOpenAccessHub(Provider):
    def query(
        self,
        satellite: "Satellite",
        processing_level: enums.ProcessingLevel,
        roi: shapely.geometry.base.BaseGeometry,
        start_date: datetime.datetime,
        end_date: datetime.datetime,
        cloud_cover: Optional[int] = None,
    ) -> List[Dict[Any, Any]]:
        """Query the Copernicus Open Access Hub for data"""

        # Check that the provider's credentials that are needed are set
        self._check_credentials_exist()

        try:
            api = SentinelAPIExtended(
                credentials.SCIHUB_USERNAME, credentials.SCIHUB_PASSWORD
            )
        except sentinelsat.UnauthorizedError as e:
            logger.error(f"ASF authentication error: {e}")
            return []

        # If cloud cover percentage is 100, set to None for API
        if cloud_cover == 100:
            cloud_cover = None

        # Variable to combine all CommonSearchResult objects into one
        all_products = []

        for product_type in self.products.get(
            (satellite.name, processing_level), []
        ):
            # We do a list here because in some cases, a satellite may have
            # multiple product types for a given processing level (sentinel 3
            # has two product types for L2)
            logger.info(f"Product type: {product_type}")
            if not product_type:
                raise ValueError(
                    f"Processing level {processing_level.value} not supported "
                    f"by Copernicus Open Access Hub for satellite "
                    f"{satellite.name}. Available processing levels: "
                    f"{self.products}"
                )

            try:
                products = api.query(
                    roi.wkt,
                    producttype=product_type,
                    cloudcoverpercentage=(0, cloud_cover)
                    if cloud_cover
                    else None,
                    date=(start_date, end_date),
                )
                # Translate the results to a common format
                products = self.translate_search_results(products)

                all_products += products
            except sentinelsat.UnauthorizedError as e:
                logger.error(f"SentinelSat authentication error: {e}")
                return []

        return all_products

    @tenacity.retry(
        retry=tenacity.retry_if_exception_type(
            sentinelsat.exceptions.ServerError
        ),
        stop=tenacity.stop_after_attempt(
            core.config.constants.MAX_DOWNLOAD_ATTEMPTS
        ),
        wait=tenacity.wait_fixed(90),  # API rate limit is 90req/60s
        reraise=True,
    )
    def download_many(
        self,
        search_results: List[CommonSearchResult],
        download_dir: str,
        overwrite: bool = False,
        processes: int = core.config.constants.PARRALLEL_PROCESSES_DEFAULT,
        *,
        max_attempts: int = core.config.constants.MAX_DOWNLOAD_ATTEMPTS,
    ) -> None:
        """Using the search results from query(), download the data

        Function will retry if the API returns a 500 error. This can happen
        when the API is under heavy load or when the API rate limit is exceeded
        which is defined at time of writing to be 90 requests per minute.

        Parameters
        ----------
        search_results : List[str]
            The search results
        download_dir : str
            The directory to download the data to
        overwrite : bool
            Whether to overwrite existing files
        processes : int
            The number of processes to use for downloading

        Returns
        -------
        None
        """
        # Check that the provider's credentials that are needed are set
        self._check_credentials_exist()

        logger.addHandler(sentinelsat.SentinelAPI.logger)

        # Convert the search results to a list of product ids
        product_ids = [result.product_id for result in search_results]

        api = SentinelAPIExtended(
            credentials.SCIHUB_USERNAME, credentials.SCIHUB_PASSWORD
        )
        api.download_all(
            product_ids,
            directory_path=download_dir,
            n_concurrent_dl=processes,
            checksum=True,
            max_attempts=max_attempts,
            overwrite=overwrite,
        )

    def translate_search_results(
        self, provider_search_results: Dict[Any, Any]
    ) -> List[CommonSearchResult]:
        """Translate search results from a provider to a common format

        Parameters
        ----------
        provider_search_results : Dict[Any, Any]
            The search results from the provider

            Returns
            -------
            List[CommonSearchResult]
                The search results in a common format
        """

        common_results = []
        for id, props in provider_search_results.items():
            # Get the satellite and processing level from reversed mapping of
            # the provider's "products" dictionary
            sat, level = self._products_reversed[props["producttype"]]

            common_results.append(
                CommonSearchResult(
                    geometry=props.get("footprint"),
                    product_id=id,
                    link=props.get("link_alternative"),
                    identifier=props.get("identifier"),
                    filename=props.get("filename"),
                    size=props.get("size"),
                    time=props.get("beginposition"),
                    cloud_cover_percentage=props.get("cloudcoverpercentage"),
                    processing_level=level,
                    satellite=sat,
                )
            )

        return common_results


copernicus_scihub: CopernicusOpenAccessHub = CopernicusOpenAccessHub(
    name="scihub",
    description="Copernicus Open Access Hub",
    uri="https://scihub.copernicus.eu",
    products={
        (enums.Satellite.SENTINEL1, enums.ProcessingLevel.L1): ["GRD"],
        (enums.Satellite.SENTINEL2, enums.ProcessingLevel.L1C): ["S2MSI1C"],
        (enums.Satellite.SENTINEL2, enums.ProcessingLevel.L2A): ["S2MSI2A"],
        (enums.Satellite.SENTINEL3, enums.ProcessingLevel.L1): ["OL_1_EFR___"],
        (enums.Satellite.SENTINEL3, enums.ProcessingLevel.L2): [
            "OL_2_LFR___",
            "OL_2_WFR___",
        ],
    },
    credentials_required=["SCIHUB_USERNAME", "SCIHUB_PASSWORD"],
)



""" Functions from the asf_search library, modified to allow overwriting of
existing files.

As the search function creates ASFProducts (where the download function belongs
to) and thus is not bound to a class, it is necessary to override several
search functions to override the ASFProduct class.

The ASFProduct class is extended to override the download function, and all
functions are given an extra argument to accept the overwrite flag.

Full repository: https://github.com/asfadmin/Discovery-asf_search

"""

import os.path
import urllib.parse
import warnings
import tqdm
from earth_extractor import core
from asf_search.download.download import _try_get_response
from asf_search.exceptions import ASFDownloadError
import logging
from multiprocessing import Pool
from typing import Generator, Union, Iterable, Tuple
from copy import copy
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_fixed,
)
import datetime
from asf_search.ASFSearchResults import ASFSearchResults
from asf_search.ASFSearchOptions import ASFSearchOptions
from asf_search.CMR import build_subqueries, translate_opts
from asf_search.ASFSession import ASFSession
from asf_search.ASFProduct import ASFProduct
from asf_search.exceptions import (
    ASFSearchError,
    CMRIncompleteError,
)
from concurrent.futures import ThreadPoolExecutor, as_completed
from asf_search.constants import INTERNAL
from asf_search.search.error_reporting import report_search_error
from asf_search.search.search_generator import (
    preprocess_opts,
    get_page,
    process_page,
)
from asf_search.download.file_download_type import FileDownloadType

# We may as well use our internal logger in this case if we have such freedom
logger = logging.getLogger(__name__)
logger.setLevel(core.config.constants.LOGLEVEL_MODULE_DEFAULT)

""" The following method and class are overriden to support the overwrite
    argument. Futher below are the functions that this overridden class
    is injected into
"""


def download_url(
    url: str,
    path: str,
    filename: str = None,
    session: ASFSession = None,
    overwrite: bool = False,
) -> None:
    """
    Downloads a product from the specified URL to the specified location and (optional) filename.

    :param url: URL from which to download
    :param path: Local path in which to save the product
    :param filename: Optional filename to be used, extracted from the URL by default
    :param session: The session to use, in most cases should be authenticated beforehand
    :param overwrite: Whether to overwrite existing files.
    :return:
    """

    if filename is None:
        filename = os.path.split(urllib.parse.urlparse(url).path)[1]

    if not os.path.isdir(path):
        raise ASFDownloadError(
            f"Error downloading {url}: directory not found: {path}"
        )

    # Allow overwriting by modifying the operation of this conditional
    if os.path.isfile(os.path.join(path, filename)):
        logger.info(
            f"File already exists: {os.path.join(filename)} "
            f"... {'Overwriting' if overwrite else 'Skipping'}"
        )

        if not overwrite:  # Don't download
            return

    if session is None:
        session = ASFSession()

    response = _try_get_response(session=session, url=url)

    # Improve download process with progress bar
    total_size = int(response.headers.get("content-length", -1))

    with open(os.path.join(path, filename), "wb") as dest:
        with tqdm.tqdm(
            total=total_size,
            desc=url,
            unit="iB",
            unit_scale=True,
            unit_divisor=1024,
        ) as bar:
            for chunk in response.iter_content(chunk_size=8192):
                if chunk:
                    size = dest.write(chunk)
                    bar.update(size)


class ASFProductExtended(ASFProduct):
    # Extend the class to override the download function
    def download(
        self,
        path: str,
        filename: str = None,
        session: ASFSession = None,
        fileType=FileDownloadType.DEFAULT_FILE,
        overwrite: bool = False,  # @evanjt Added to support overwriting
    ) -> None:
        """
        Downloads this product to the specified path and optional filename.

        :param path: The directory into which this product should be downloaded.
        :param filename: Optional filename to use instead of the original filename of this product.
        :param session: The session to use, defaults to the one used to find the results.
        :param overwrite: Whether to overwrite existing files.
        :return: None
        """

        default_filename = self.properties["fileName"]

        if filename is not None:
            multiple_files = (
                fileType == FileDownloadType.ADDITIONAL_FILES
                and len(self.properties["additionalUrls"]) > 1
            ) or fileType == FileDownloadType.ALL_FILES
            if multiple_files:
                warnings.warn(
                    f'Attempting to download multiple files for product, ignoring user provided filename argument "{filename}", using default.'
                )
            else:
                default_filename = filename

        if session is None:
            session = self.session

        urls = []

        def get_additional_urls():
            output = []
            base_filename = ".".join(default_filename.split(".")[:-1])
            for url in self.properties["additionalUrls"]:
                extension = url.split(".")[-1]
                urls.append((f"{base_filename}.{extension}", url))

            return output

        if fileType == FileDownloadType.DEFAULT_FILE:
            urls.append((default_filename, self.properties["url"]))
        elif fileType == FileDownloadType.ADDITIONAL_FILES:
            urls.extend(get_additional_urls())
        elif fileType == FileDownloadType.ALL_FILES:
            urls.append((default_filename, self.properties["url"]))
            urls.extend(get_additional_urls())
        else:
            raise ValueError(
                "Invalid FileDownloadType provided, the valid types are 'DEFAULT_FILE', 'ADDITIONAL_FILES', and 'ALL_FILES'"
            )

        for filename, url in urls:
            download_url(
                url=url,
                path=path,
                filename=filename,
                session=session,
                overwrite=overwrite,  # @evanjt Added to support overwriting
            )


""" As the search function creates a list of ASFProducts in its return,
    the following functions are required to inject the ASFProductExtended class
    containing the overridden download function.
"""


def search_generator(
    absoluteOrbit: Union[
        int, Tuple[int, int], Iterable[Union[int, Tuple[int, int]]]
    ] = None,
    asfFrame: Union[
        int, Tuple[int, int], Iterable[Union[int, Tuple[int, int]]]
    ] = None,
    beamMode: Union[str, Iterable[str]] = None,
    beamSwath: Union[str, Iterable[str]] = None,
    campaign: Union[str, Iterable[str]] = None,
    maxDoppler: float = None,
    minDoppler: float = None,
    end: Union[datetime.datetime, str] = None,
    maxFaradayRotation: float = None,
    minFaradayRotation: float = None,
    flightDirection: str = None,
    flightLine: str = None,
    frame: Union[
        int, Tuple[int, int], Iterable[Union[int, Tuple[int, int]]]
    ] = None,
    granule_list: Union[str, Iterable[str]] = None,
    groupID: Union[str, Iterable[str]] = None,
    insarStackId: str = None,
    instrument: Union[str, Iterable[str]] = None,
    intersectsWith: str = None,
    lookDirection: Union[str, Iterable[str]] = None,
    offNadirAngle: Union[
        float, Tuple[float, float], Iterable[Union[float, Tuple[float, float]]]
    ] = None,
    platform: Union[str, Iterable[str]] = None,
    polarization: Union[str, Iterable[str]] = None,
    processingDate: Union[datetime.datetime, str] = None,
    processingLevel: Union[str, Iterable[str]] = None,
    product_list: Union[str, Iterable[str]] = None,
    relativeOrbit: Union[
        int, Tuple[int, int], Iterable[Union[int, Tuple[int, int]]]
    ] = None,
    season: Tuple[int, int] = None,
    start: Union[datetime.datetime, str] = None,
    absoluteBurstID: Union[int, Iterable[int]] = None,
    relativeBurstID: Union[int, Iterable[int]] = None,
    fullBurstID: Union[str, Iterable[str]] = None,
    collections: Union[str, Iterable[str]] = None,
    temporalBaselineDays: Union[str, Iterable[str]] = None,
    maxResults: int = None,
    opts: ASFSearchOptions = None,
) -> Generator[ASFSearchResults, None, None]:
    # Create a kwargs dict, that's all of the 'not None' items, and merge it with opts:
    kwargs = locals()
    opts = ASFSearchOptions() if kwargs["opts"] is None else copy(opts)
    del kwargs["opts"]

    kwargs = dict((k, v) for k, v in kwargs.items() if v is not None)
    kw_opts = ASFSearchOptions(**kwargs)

    # Anything passed in as kwargs has priority over anything in opts:
    opts.merge_args(**dict(kw_opts))

    maxResults = opts.pop("maxResults", None)

    if maxResults is not None and (
        getattr(opts, "granule_list", False)
        or getattr(opts, "product_list", False)
    ):
        raise ValueError(
            "Cannot use maxResults along with product_list/granule_list."
        )

    preprocess_opts(opts)

    url = "/".join(
        s.strip("/")
        for s in [f"https://{opts.host}", f"{INTERNAL.CMR_GRANULE_PATH}"]
    )
    total = 0

    queries = build_subqueries(opts)
    for query in queries:
        translated_opts = translate_opts(query)
        cmr_search_after_header = ""
        subquery_count = 0

        while cmr_search_after_header is not None:
            try:
                (
                    items,
                    subquery_max_results,
                    cmr_search_after_header,
                ) = query_cmr(  # @evanjt: Edited to allow local import
                    opts.session, url, translated_opts, subquery_count
                )
            except (ASFSearchError, CMRIncompleteError) as e:
                message = str(e)
                logging.error(message)
                report_search_error(query, message)
                opts.session.headers.pop("CMR-Search-After", None)
                return

            opts.session.headers.update(
                {"CMR-Search-After": cmr_search_after_header}
            )
            last_page = process_page(
                items,
                maxResults,
                subquery_max_results,
                total,
                subquery_count,
                opts,
            )
            subquery_count += len(last_page)
            total += len(last_page)
            last_page.searchComplete = (
                subquery_count == subquery_max_results or total == maxResults
            )
            yield last_page

            if last_page.searchComplete:
                if (
                    total == maxResults
                ):  # the user has as many results as they wanted
                    opts.session.headers.pop("CMR-Search-After", None)
                    return
                else:  # or we've gotten all possible results for this subquery
                    cmr_search_after_header = None

        opts.session.headers.pop("CMR-Search-After", None)


@retry(
    reraise=True,
    retry=retry_if_exception_type(CMRIncompleteError),
    wait=wait_fixed(2),
    stop=stop_after_attempt(3),
)
def query_cmr(
    session: ASFSession, url: str, translated_opts: dict, sub_query_count: int
):
    response = get_page(
        session=session, url=url, translated_opts=translated_opts
    )

    items = [
        ASFProductExtended(  # @evanjt Modified here to support overwriting
            f, session=session
        )
        for f in response.json()["items"]
    ]
    hits: int = response.json()[
        "hits"
    ]  # total count of products given search opts

    # sometimes CMR returns results with the wrong page size
    if (
        len(items) != INTERNAL.CMR_PAGE_SIZE
        and len(items) + sub_query_count < hits
    ):
        raise CMRIncompleteError(
            f"CMR returned page of incomplete results. Expected {min(INTERNAL.CMR_PAGE_SIZE, hits - sub_query_count)} results, got {len(items)}"
        )

    return items, hits, response.headers.get("CMR-Search-After", None)


def search(
    absoluteOrbit: Union[
        int, Tuple[int, int], Iterable[Union[int, Tuple[int, int]]]
    ] = None,
    asfFrame: Union[
        int, Tuple[int, int], Iterable[Union[int, Tuple[int, int]]]
    ] = None,
    beamMode: Union[str, Iterable[str]] = None,
    beamSwath: Union[str, Iterable[str]] = None,
    campaign: Union[str, Iterable[str]] = None,
    maxDoppler: float = None,
    minDoppler: float = None,
    end: Union[datetime.datetime, str] = None,
    maxFaradayRotation: float = None,
    minFaradayRotation: float = None,
    flightDirection: str = None,
    flightLine: str = None,
    frame: Union[
        int, Tuple[int, int], Iterable[Union[int, Tuple[int, int]]]
    ] = None,
    granule_list: Union[str, Iterable[str]] = None,
    groupID: Union[str, Iterable[str]] = None,
    insarStackId: str = None,
    instrument: Union[str, Iterable[str]] = None,
    intersectsWith: str = None,
    lookDirection: Union[str, Iterable[str]] = None,
    offNadirAngle: Union[
        float, Tuple[float, float], Iterable[Union[float, Tuple[float, float]]]
    ] = None,
    platform: Union[str, Iterable[str]] = None,
    polarization: Union[str, Iterable[str]] = None,
    processingDate: Union[datetime.datetime, str] = None,
    processingLevel: Union[str, Iterable[str]] = None,
    product_list: Union[str, Iterable[str]] = None,
    relativeOrbit: Union[
        int, Tuple[int, int], Iterable[Union[int, Tuple[int, int]]]
    ] = None,
    season: Tuple[int, int] = None,
    start: Union[datetime.datetime, str] = None,
    absoluteBurstID: Union[int, Iterable[int]] = None,
    relativeBurstID: Union[int, Iterable[int]] = None,
    fullBurstID: Union[str, Iterable[str]] = None,
    collections: Union[str, Iterable[str]] = None,
    temporalBaselineDays: Union[str, Iterable[str]] = None,
    maxResults: int = None,
    opts: ASFSearchOptions = None,
) -> ASFSearchResults:
    """
    Performs a generic search using the ASF SearchAPI. Accepts a number of search parameters, and/or an ASFSearchOptions object. If an ASFSearchOptions object is provided as well as other specific parameters, the two sets of options will be merged, preferring the specific keyword arguments.

    :param absoluteOrbit: For ALOS, ERS-1, ERS-2, JERS-1, and RADARSAT-1, Sentinel-1A, Sentinel-1B this value corresponds to the orbit count within the orbit cycle. For UAVSAR it is the Flight ID.
    :param asfFrame: This is primarily an ASF / JAXA frame reference. However, some platforms use other conventions. See ‘frame’ for ESA-centric frame searches.
    :param beamMode: The beam mode used to acquire the data.
    :param beamSwath: Encompasses a look angle and beam mode.
    :param campaign: For UAVSAR and AIRSAR data collections only. Search by general location, site description, or data grouping as supplied by flight agency or project.
    :param maxDoppler: Doppler provides an indication of how much the look direction deviates from the ideal perpendicular flight direction acquisition.
    :param minDoppler: Doppler provides an indication of how much the look direction deviates from the ideal perpendicular flight direction acquisition.
    :param end: End date of data acquisition. Supports timestamps as well as natural language such as "3 weeks ago"
    :param maxFaradayRotation: Rotation of the polarization plane of the radar signal impacts imagery, as HH and HV signals become mixed.
    :param minFaradayRotation: Rotation of the polarization plane of the radar signal impacts imagery, as HH and HV signals become mixed.
    :param flightDirection: Satellite orbit direction during data acquisition
    :param flightLine: Specify a flightline for UAVSAR or AIRSAR.
    :param frame: ESA-referenced frames are offered to give users a universal framing convention. Each ESA frame has a corresponding ASF frame assigned. See also: asfframe
    :param granule_list: List of specific granules. Search results may include several products per granule name.
    :param groupID: Identifier used to find products considered to be of the same scene but having different granule names
    :param insarStackId: Identifier used to find products of the same InSAR stack
    :param instrument: The instrument used to acquire the data. See also: platform
    :param intersectsWith: Search by polygon, linestring, or point defined in 2D Well-Known Text (WKT)
    :param lookDirection: Left or right look direction during data acquisition
    :param offNadirAngle: Off-nadir angles for ALOS PALSAR
    :param platform: Remote sensing platform that acquired the data. Platforms that work together, such as Sentinel-1A/1B and ERS-1/2 have multi-platform aliases available. See also: instrument
    :param polarization: A property of SAR electromagnetic waves that can be used to extract meaningful information about surface properties of the earth.
    :param processingDate: Used to find data that has been processed at ASF since a given time and date. Supports timestamps as well as natural language such as "3 weeks ago"
    :param processingLevel: Level to which the data has been processed
    :param product_list: List of specific products. Guaranteed to be at most one product per product name.
    :param relativeOrbit: Path or track of satellite during data acquisition. For UAVSAR it is the Line ID.
    :param season: Start and end day of year for desired seasonal range. This option is used in conjunction with start/end to specify a seasonal range within an overall date range.
    :param start: Start date of data acquisition. Supports timestamps as well as natural language such as "3 weeks ago"
    :param collections: List of collections (concept-ids) to limit search to
    :param temporalBaselineDays: List of temporal baselines, used for Sentinel-1 Interferogram (BETA)
    :param maxResults: The maximum number of results to be returned by the search
    :param opts: An ASFSearchOptions object describing the search parameters to be used. Search parameters specified outside this object will override in event of a conflict.

    :return: ASFSearchResults(list) of search results
    """
    kwargs = locals()
    data = dict(
        (k, v)
        for k, v in kwargs.items()
        if k not in ["host", "opts"] and v is not None
    )

    opts = ASFSearchOptions() if opts is None else copy(opts)
    opts.merge_args(**data)

    results = ASFSearchResultsExtended([])

    # The last page will be marked as complete if results sucessful
    for page in search_generator(  # @evanjt: Use local function for overwrite
        opts=opts
    ):
        results.extend(page)
        results.searchComplete = page.searchComplete
        results.searchOptions = page.searchOptions

    results.sort(
        key=lambda p: (p.properties["stopTime"], p.properties["fileID"]),
        reverse=True,
    )

    return results


def granule_search(
    granule_list: Iterable[str],
    opts: ASFSearchOptions = None,
) -> ASFSearchResults:
    """
    Performs a granule name search using the ASF SearchAPI

    :param granule_list: List of specific granules. Results may include several products per granule name.
    :param opts: An ASFSearchOptions object describing the search parameters to be used. Search parameters specified outside this object will override in event of a conflict.

    :return: ASFSearchResults(list) of search results
    """

    opts = ASFSearchOptions() if opts is None else copy(opts)

    opts.merge_args(granule_list=granule_list)

    return search(opts=opts)  # @evanjt: Use local function for overwrite


class ASFSearchResultsExtended(ASFSearchResults):
    def download(
        self,
        path: str,
        session: ASFSession = None,
        processes: int = 1,
        fileType=FileDownloadType.DEFAULT_FILE,
        overwrite: bool = False,
    ) -> None:
        """
        Iterates over each ASFProduct and downloads them to the specified path.

        :param path: The directory into which the products should be downloaded.
        :param session: The session to use. Defaults to the session used to fetch the results, or a new one if none was used.
        :param processes: Number of download processes to use. Defaults to 1 (i.e. sequential download)
        :param overwrite: Whether to overwrite existing files.

        :return: None
        """
        logger.info(
            f"Started downloading ASFSearchResults of size {len(self)}."
        )
        if processes == 1:
            for product in self:
                product.download(
                    path=path,
                    session=session,
                    fileType=fileType,
                    overwrite=overwrite,
                )
        else:
            logger.info(f"Using {processes} threads - starting up pool.")

            with ThreadPoolExecutor(max_workers=processes) as executor:
                futures = {
                    executor.submit(
                        _download_product,
                        (product, path, session, fileType, overwrite),
                    )
                    for product in self
                }
                for future in as_completed(futures):
                    try:
                        result = future.result()
                        logger.debug(
                            "Downloaded file successfully (threading): "
                            f"({result})"
                        )
                    except Exception as e:
                        logger.error(
                            f"ASF downloading generated an exception: {e}"
                        )

    def raise_if_incomplete(self) -> None:
        if not self.searchComplete:
            msg = "Results are incomplete due to a search error. See logging for more details. (ASFSearchResults.raise_if_incomplete called)"
            logger.error(msg)
            raise ASFSearchError(msg)


def _download_product(args) -> None:
    product, path, session, fileType, overwrite = args
    product.download(
        path=path, session=session, fileType=fileType, overwrite=overwrite
    )



import keyring
import typer
import logging
import sys
from rich.console import Console
from rich.table import Table
from pydantic import BaseSettings, root_validator
from typing import Dict, Optional
from earth_extractor import core
import jwt
import datetime
from functools import lru_cache

logger = logging.getLogger(__name__)
logger.setLevel(core.config.constants.LOGLEVEL_MODULE_DEFAULT)


class Credentials(BaseSettings):
    # User tokens
    SCIHUB_USERNAME: Optional[str] = None
    SCIHUB_PASSWORD: Optional[str] = None

    # NASA (For Alaska Satellite Facility and NASA Common Metadata Repository)
    NASA_TOKEN: Optional[str] = None

    # Sinergise Sentinel Hub
    SINERGISE_CLIENT_ID: Optional[str] = None
    SINERGISE_CLIENT_SECRET: Optional[str] = None

    @root_validator
    def populate_credentials_from_keyring(
        cls, values: Dict[str, Optional[str]]
    ) -> Dict[str, Optional[str]]:
        """Populate the credentials from the keyring"""
        for key in values.keys():
            values[key] = keyring.get_password(
                core.config.constants.KEYRING_ID, key
            )
        if "pytest" in sys.modules:  # Populate fake credentials for unit tests
            return {
                "SCIHUB_USERNAME": "test",
                "SCIHUB_PASSWORD": "test",
                "NASA_TOKEN": jwt.encode(  # JSON Web token
                    {
                        "some": "test",
                        "iat": datetime.datetime.utcnow(),
                        "exp": datetime.datetime.utcnow()
                        + datetime.timedelta(days=1),
                    },
                    "secret",
                    algorithm="HS256",
                ),
                "SINERGISE_CLIENT_ID": "test",
                "SINERGISE_CLIENT_SECRET": "test",
            }

        return values


def show_credential_list(show_secret=False) -> None:
    """Lists the credential keys in a table to the console

    Parameters
    ----------
    show_secret : bool
        If True, the secret values are shown in the table. If False, the
        secret values are hidden. Default is False.

    Returns
    -------
    None
    """

    console = Console()
    credentials = get_credentials()
    table = Table(title="Credentials")
    table.add_column("Credential key", justify="left")

    if show_secret:
        table.add_column("Value", justify="left")
    else:
        table.add_column("Value is set", justify="center")

    for cred_key in credentials.__fields__:
        if show_secret:
            value = getattr(credentials, cred_key)
        else:
            if getattr(credentials, cred_key) is not None:
                value = "[green]Yes[/green]"
            else:
                value = "[red]No[/red]"
        table.add_row(cred_key, value)

    console.print(table)


def set_one_credential(
    key,
    hide_prompt=core.config.constants.HIDE_PASSWORD_PROMPT,
) -> None:
    """Set a single credential key in the keyring

    Parameters
    ----------
    key : str
        The key to set
    hide_prompt : bool
        If True, the prompt is hidden. If False, the prompt is shown.

    """

    if key not in get_credentials().__fields__:
        raise ValueError(f"Key '{key}' does not exist")

    secret = getattr(get_credentials(), key)

    new_secret = typer.prompt(
        key, default="" if secret is None else secret, hide_input=hide_prompt
    )

    if new_secret == "":
        # Don't store '' in the keyring in case there are any, just delete
        if secret == "":
            keyring.delete_password(core.config.constants.KEYRING_ID, key)
    else:
        keyring.set_password(core.config.constants.KEYRING_ID, key, new_secret)


def set_all_credentials() -> None:
    """Set all credential keys in the keyring"""

    for cred_key in get_credentials().__fields__:
        set_one_credential(cred_key)


def delete_credential(key) -> None:
    if key not in get_credentials().__fields__:
        raise ValueError(f"Key '{key}' does not exist")

    keyring.delete_password(core.config.constants.KEYRING_ID, key)


@lru_cache
def get_credentials() -> Credentials:
    return Credentials()
